//////////////////////////////////////////////////////////////
///                                                        ///
/// dvb_front.c: DVB frontend functions of dtv             ///
///                                                        ///
//////////////////////////////////////////////////////////////

// $Revision: 1.18 $
// $Date: 2001/06/14 17:03:51 $
// $Author: hakenes $
//
//   (C) 2001 Rolf Hakenes <hakenes@hippomi.de>, under the GNU GPL.
//
// dtv is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// dtv is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You may have received a copy of the GNU General Public License
// along with dtv; see the file COPYING.  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.

#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>

#include <ost/dmx.h>
#include <ost/frontend.h>
#include <ost/sec.h>
#include <ost/video.h>

#include <libsi.h>

#include "definitions.h"
#include "structures.h"
#include "prototypes.h"


static char sbuf[4096];

static struct LIST *TmpChanList = NULL;
static struct MEM_CHUNK *TmpChanListRemember = NULL;

struct OutputDevice *CurrentOutputDevice = NULL;

int openDvbDevice (struct OutputDevice *OutputDevice)
{
   if (!xName (OutputDevice)) return (-1);

   sprintf(sbuf, "/dev/ost/demux%s", xName(OutputDevice));
   if ((OutputDevice->audioDev=open(sbuf, O_RDWR | O_NONBLOCK)) < 0)
    {  logPrintf (0, "cannot open %s (%s)\n", sbuf, strerror (errno));
       return (-1); }

   sprintf(sbuf, "/dev/ost/demux%s", xName(OutputDevice));
   if ((OutputDevice->videoDev=open(sbuf, O_RDWR | O_NONBLOCK)) < 0)
    {  logPrintf (0, "cannot open %s (%s)\n", sbuf, strerror (errno));
       return (-1); }

   if (GetGlobalOption ("disableOsd") != 0) return (0);

#ifdef USE_DVB_0_8
   sprintf(sbuf, "/dev/video%s", xName(OutputDevice));
#else
   sprintf(sbuf, "/dev/ost/osd%s", xName(OutputDevice));
#endif
   if ((OutputDevice->osdDev = open(sbuf, O_RDWR)) < 0)
    {  logPrintf (0, "cannot open %s (%s)\n", sbuf, strerror (errno));
       return (-1); }

   return (openOsdDevice (OutputDevice));
}


void closeDvbDevice (struct OutputDevice *OutputDevice)
{
   if (OutputDevice->osdDev)
   {
      closeOsdDevice (OutputDevice);
      close (OutputDevice->osdDev);
   }

   if (!OutputDevice->audioDev)
   {
      ioctl(OutputDevice->audioDev, DMX_STOP, 0);
      close (OutputDevice->audioDev);
   }
   if (!OutputDevice->videoDev)
   {
      ioctl(OutputDevice->videoDev, DMX_STOP, 0);
      close (OutputDevice->videoDev);
   }

   return;
}


int setDvbChannel (struct OutputDevice *OutputDevice, struct Channel *Channel)
{
   int                          fd_frontend, fd_sec, newFrequency;
   struct dmxPesFilterParams    pesFilterParams;
   int                          i = 2, freq, count, res;
   feStatus                     stat = 0;

   if (!OutputDevice || !Channel)
    {  logPrintf (0, "setDvbChannel called with invalid parameters\n");
       return (-1); }

   if (OutputDevice->audioDev) ioctl(OutputDevice->audioDev, DMX_STOP, 0);
   close (OutputDevice->audioDev);
   if (OutputDevice->videoDev) ioctl(OutputDevice->videoDev, DMX_STOP, 0);
   close (OutputDevice->videoDev);

   if (Channel->Polarisation == 'c' || Channel->Polarisation == 'C')
   {
      struct qamParameters qam;
      struct qamEvent event;

      sprintf(sbuf, "/dev/ost/qamfe%s", xName(OutputDevice));
      if ((fd_frontend=open(sbuf, O_RDWR)) < 0)
       {  logPrintf (0, "cannot open %s (%s)\n", sbuf, strerror (errno));
          return (-1); }

      qam.FEC_inner = 0 /* FEC_AUTO */;
      qam.SymbolRate = Channel->SymbolRate * 1000;
      qam.QAM = 2 /* QAM_64 */;
      qam.Frequency = Channel->Frequency * 1000; /* this is due to the fact,
                                                    that frequencies for QAM
                                                    have to be specified in Hz,
                                                    for QPSK in kHz...
                                                    why the hell is that ?? */
      if (ioctl(fd_frontend, QAM_TUNE, &qam) <0)
       { logPrintf (0, "setfront qamfe error (%s)\n", strerror (errno));
         close (fd_frontend); return (-1); }

      if ((res = ioctl(fd_frontend, QAM_GET_EVENT, &event)) == -EBUFFEROVERFLOW)
         res = ioctl(fd_frontend, QAM_GET_EVENT, &event);
      if (res < 0)
       { logPrintf (0, "get event qamfe error (%s)\n", strerror (errno));
         close (fd_frontend); return (-1); }

      if (event.type != FE_COMPLETION_EV)
       { logPrintf (0, "could not tune to %d/%c/%d\n", Channel->Frequency,
            Channel->Polarisation, Channel->SymbolRate);
         close (fd_frontend); return (-1); }

      newFrequency = event.u.completionEvent.Frequency/1000; /* see above */
      if (newFrequency != Channel->Frequency)
         logPrintf (4, "frequency adjusted from %ld to %ld\n",
             Channel->Frequency, newFrequency);


      close (fd_frontend);
   }
   else
   {
      struct qpskParameters qpsk;
      struct qpskEvent event;
      struct secCommand scmd;
      struct secCmdSequence scmds;
      int DiseqcNr, LnbLsof, LnbFreqLow, LnbFreqHigh;

      if (Channel->Satellite)
      {
         DiseqcNr = Channel->Satellite->DiseqcNr;
         LnbLsof = Channel->Satellite->LnbLsof;
         LnbFreqLow = Channel->Satellite->LnbFreqLow;
         LnbFreqHigh = Channel->Satellite->LnbFreqHigh;
      }
      else
      {
         DiseqcNr = GetGlobalOption ("dvbDefaultDiseqcNr");
         LnbLsof = GetGlobalOption ("dvbDefaultLnbLsof");
         LnbFreqLow = GetGlobalOption ("dvbDefaultLnbFreqLow");
         LnbFreqHigh = GetGlobalOption ("dvbDefaultLnbFreqHigh");
      }
      LnbLsof *= 1000; LnbFreqLow*= 1000; LnbFreqHigh *= 1000;

      sprintf(sbuf, "/dev/ost/qpskfe%s", xName(OutputDevice));
      if ((fd_frontend=open(sbuf, O_RDWR)) < 0)
       {  logPrintf (0, "cannot open %s (%s)\n", sbuf, strerror (errno));
          return (-1); }

      sprintf(sbuf, "/dev/ost/sec%s", xName(OutputDevice));
      if ((fd_sec=open(sbuf, O_RDWR)) < 0)
       {  logPrintf (0, "cannot open %s (%s)\n", sbuf, strerror (errno));
          close (fd_frontend); return (-1); }

      qpsk.FEC_inner = 0 /* FEC_AUTO */;
      qpsk.SymbolRate = Channel->SymbolRate * 1000;

      if (Channel->Frequency < LnbLsof)
      {
         qpsk.iFrequency = Channel->Frequency - LnbFreqLow;
         scmds.continuousTone = SEC_TONE_OFF;
      }
      else
      {
         qpsk.iFrequency = Channel->Frequency - LnbFreqHigh;
         scmds.continuousTone = SEC_TONE_ON;
      }
      if (Channel->Polarisation == 'h' || Channel->Polarisation == 'H')
         scmds.voltage = SEC_VOLTAGE_18;
      else scmds.voltage = SEC_VOLTAGE_13;

      /* set_diseqc_nb(scmd, scmds, nr); */
      scmd.type=0;
      scmd.u.diseqc.addr = 0x10;
      scmd.u.diseqc.cmd = 0x38;
      scmd.u.diseqc.numParams = 1;
      scmd.u.diseqc.params[0] = 0xF0 | ((DiseqcNr * 4) & 0x0F) |
        (scmds.continuousTone == SEC_TONE_ON ? 1 : 0) |
        (scmds.voltage==SEC_VOLTAGE_18 ? 2 : 0);

      scmds.miniCommand = SEC_MINI_NONE;
      scmds.numCommands=1;
      scmds.commands=&scmd;

      freq = qpsk.iFrequency;

      if (ioctl(fd_sec, SEC_SEND_SEQUENCE, &scmds) < 0)
       { logPrintf (0, "setfront sec error (%s)\n", strerror (errno));
         close (fd_sec); close (fd_frontend); return (-1); }
   
      if (ioctl(fd_frontend, QPSK_TUNE, &qpsk) <0)
       { logPrintf (0, "setfront qpskfe error (%s)\n", strerror (errno));
         close (fd_sec); close (fd_frontend); return (-1); }
      if ((res = ioctl(fd_frontend, QPSK_GET_EVENT, &event))==-EBUFFEROVERFLOW)
            res = ioctl(fd_frontend, QPSK_GET_EVENT, &event);
      if (res < 0)
       { logPrintf (0, "get event qpskfe error (%s)\n", strerror (errno));
         close (fd_sec); close (fd_frontend); return (-1); }

      if (event.type != FE_COMPLETION_EV)
       { logPrintf (0, "could not tune to %d/%c/%d\n", Channel->Frequency,
            Channel->Polarisation, Channel->SymbolRate);
         close (fd_sec); close (fd_frontend); return (-1); }

      newFrequency = Channel->Frequency +
                        (event.u.completionEvent.iFrequency/1000 - freq);
      if (newFrequency != Channel->Frequency)
         logPrintf (4, "frequency adjusted from %ld to %ld\n",
             Channel->Frequency, newFrequency);

      close (fd_sec);
      close (fd_frontend);
   }

   sprintf(sbuf, "/dev/ost/demux%s", xName(OutputDevice));
   if ((OutputDevice->videoDev=open(sbuf, O_RDWR | O_NONBLOCK)) < 0)
    {  logPrintf (0, "cannot open %s (%s)\n", sbuf, strerror (errno));
       return (-1); }
   if (Channel->VideoPid!=0 &&
       Channel->VideoPid!=0xffff &
       Channel->VideoPid!=0x1fff)
   {
      pesFilterParams.pid = Channel->VideoPid;
      pesFilterParams.input = DMX_IN_FRONTEND;
      pesFilterParams.output = DMX_OUT_DECODER;
      pesFilterParams.pesType = DMX_PES_VIDEO;
      pesFilterParams.flags = DMX_IMMEDIATE_START;
      if (ioctl(OutputDevice->videoDev, DMX_SET_PES_FILTER,
           &pesFilterParams) < 0)
       { logPrintf (0, "set_vpid error (%s)\n", strerror (errno)); }
   }

   sprintf(sbuf, "/dev/ost/demux%s", xName(OutputDevice));
   if ((OutputDevice->audioDev=open(sbuf, O_RDWR | O_NONBLOCK)) < 0)
    {  logPrintf (0, "cannot open %s (%s)\n", sbuf, strerror (errno));
       return (-1); }
   if (Channel->AudioPid!=0 &&
       Channel->AudioPid!=0xffff &
       Channel->AudioPid!=0x1fff)
   {
      pesFilterParams.pid = Channel->AudioPid;
      pesFilterParams.input = DMX_IN_FRONTEND;
      pesFilterParams.output = DMX_OUT_DECODER;
      pesFilterParams.pesType = DMX_PES_AUDIO;
      pesFilterParams.flags = DMX_IMMEDIATE_START;
      if (ioctl(OutputDevice->audioDev, DMX_SET_PES_FILTER,
           &pesFilterParams) < 0)
      { logPrintf (0, "set_apid error (%s)\n", strerror (errno)); }
   }

   return (newFrequency);
}


int processDvbCommand (struct OutputDevice *OutputDevice,
                      struct Command *Command)
{
   struct Channel         *Channel, *chan, *tchan;
   struct Command         *NewCommand;
   struct VtxPage         *VtxPage;
   time_t                  ThisTime;
   struct tm              *StartTime;
   char                   *sptr, *tptr;
   int                     newFreq, nextPage, curPage, curSub;

#define ImplicitOutputID OutputDevice->Identifier

   logPrintf (7, "%s, %d, %d - %d\n", xName (OutputDevice),
            OutputDevice->Identifier, Command->OutputID, Command->Code);

   CurrentOutputDevice = OutputDevice;

   switch (Command->Code)
   {
      case COMCODE_ON:
         if (!xSize (OutputDevice->Channels))
          { logPrintf (0, "no channels defined for device %s\n",
               xName (OutputDevice)); break; }
         if (!OutputDevice->Channel) OutputDevice->Channel =
             (struct Channel *)(OutputDevice->Channels->Head);
         if (openDvbDevice (OutputDevice) < 0)
          { logPrintf (0, "unable to open device %s\n",
               xName (OutputDevice)); break; }
         OpenDependentInputDevices (IMPL_TDT);
         OutputDevice->State = OUTST_IDLE;
         gettimeofday (&(OutputDevice->TimeStamp), NULL);
      break;

      case COMCODE_OFF:
         PurgeAllCommands (Command);
         CloseDependentInputDevices (IMPL_ALL);
         closeDvbDevice (OutputDevice);
         logPrintf (0, "Device %s switched off\n", xName (OutputDevice));
         OutputDevice->State = OUTST_OFF;
         OutputDevice->TimeStamp.tv_sec = 0;
         OutputDevice->TimeStamp.tv_usec = 0;
      break;

      case COMCODE_CHANNEL_SET:
         FindChannel (OutputDevice, Channel, Command->Value);
         if (Channel) OutputDevice->Channel = Channel;
         else break;
         CloseDependentInputDevices (IMPL_VTX|IMPL_PFO|IMPL_EIT);
         if ((newFreq = setDvbChannel (OutputDevice, OutputDevice->Channel)) < 0)
         {  sprintf (sbuf, " \f%c%s\f%c\n", COL2ID (clrRed),
               screenMsgs[COULD_NOT_TUNE_CHANNEL][OutputDevice->osdLanguage],
               COL2ID (clrWhite));
            logPrintf (0, "could not tune channel\n");
            ScheduleDirectText (COMCODE_INFO_MESSAGE, sbuf); break; }
//         Channel->Frequency = newFreq;
         PurgeCommands (COMCODE_PROGRAM_INFO);
         OpenDependentInputDevices (IMPL_PFO);
         ScheduleDirectPointer (COMCODE_PROGRAM_INFO,
            OutputDevice->Channel, 5);
         gettimeofday (&(OutputDevice->TimeStamp), NULL);
      break;

      case COMCODE_CHANNEL_DOWN:
         Channel = (struct Channel *) xPred (OutputDevice->Channel);
         if (Channel)
         {
            ScheduleDirectValue (COMCODE_CHANNEL_INFO, Channel->ChannelNumber);
            PurgeCommands (COMCODE_CHANNEL_SET);
            ScheduleDirectValue (COMCODE_CHANNEL_SET, Channel->ChannelNumber);
         }
      break;

      case COMCODE_CHANNEL_UP:
         Channel = (struct Channel *) xSucc (OutputDevice->Channel);
         if (Channel)
         {
            ScheduleDirectValue (COMCODE_CHANNEL_INFO, Channel->ChannelNumber);
            PurgeCommands (COMCODE_CHANNEL_SET);
            ScheduleDirectValue (COMCODE_CHANNEL_SET, Channel->ChannelNumber);
         }
      break;

      case COMCODE_CHANNEL_INFO:
         if (Command->Value == -1) Channel = OutputDevice->Channel;
         else FindChannel (OutputDevice, Channel, Command->Value);
         ThisTime = time (NULL);
         StartTime = localtime (&ThisTime);
         osdClear (OutputDevice->osdDev);
         osdTermClear (OutputDevice->OsdTerm);
         sprintf (sbuf, " \f%c%d\t2\f%c%s\n", COL2ID (clrYellow),
            Channel ? Channel->ChannelNumber : Command->Value, COL2ID (clrWhite),
            Channel ? xName(Channel) : "");
         osdTermChannelOut (OutputDevice->OsdTerm, sbuf);
         sprintf (sbuf, "  \f%c%02d:%02d\f%c  \n", COL2ID (clrGreen),
            StartTime->tm_hour, StartTime->tm_min, COL2ID (clrWhite));
         osdTermTimeOut (OutputDevice->OsdTerm, sbuf);
         osdShow (OutputDevice->osdDev);
         logPrintf (3, "channel info %s\n", Channel ? xName(Channel) : "-");
         PurgeCommands (COMCODE_CLEAR_INFO);
         ScheduleDelayed (COMCODE_CLEAR_INFO,
            GetGlobalOption ("dvbDefaultClearDelay") * 10);
      break;

      case COMCODE_PROGRAM_INFO:
         if (Command->Pointer) Channel = (struct Channel *)Command->Pointer;
         else Channel = OutputDevice->Channel;
         if (Channel)
         {
            if  (tptr = getPresentFollowingInfo (Channel))
            {
               osdTermPresentFollowingOut (OutputDevice->OsdTerm, tptr);
               osdClearEow (OutputDevice);
               osdShow (OutputDevice->osdDev);              
               logPrintf (3, "present/following info\n");
               PurgeCommands (COMCODE_CLEAR_INFO);
               ScheduleDelayed (COMCODE_CLEAR_INFO,
                    GetGlobalOption ("dvbDefaultClearDelay") * 10);
            }
            else if (Command->Value)
               ScheduleDelayedPointer (COMCODE_PROGRAM_INFO,
                  Channel, Command->Value-1, 10);
         }
      break;

      case COMCODE_INFO_MESSAGE:
         if (xName (Command))
         {
            osdTermMessageOut (OutputDevice, xName (Command));
            osdShow (OutputDevice->osdDev);
            logPrintf (5, "info message \"%s\"\n", xName (Command));
            PurgeCommands (COMCODE_CLEAR_INFO);
            if (Command->Value)
               ScheduleDelayed (COMCODE_CLEAR_INFO, Command->Value * 10);
            else ScheduleDelayed (COMCODE_CLEAR_INFO,
                    GetGlobalOption ("dvbDefaultClearDelay") * 10);
         }
      break;

      case COMCODE_INFO_TEXT:
         if (xName (Command))
         {
            osdTermMenuOut (OutputDevice->OsdTerm, xName (Command));
            osdShow (OutputDevice->osdDev);
            logPrintf (5, "info text \"%s\"\n", xName (Command));
            PurgeCommands (COMCODE_CLEAR_INFO);
            if (Command->Value)
               ScheduleDelayed (COMCODE_CLEAR_INFO, Command->Value * 10);
            else ScheduleDelayed (COMCODE_CLEAR_INFO,
                    GetGlobalOption ("dvbDefaultClearDelay") * 10);
         }
      break;

      case COMCODE_CLEAR_INFO:
         osdHide (OutputDevice->osdDev);
         osdClear (OutputDevice->osdDev);
         osdTermClear (OutputDevice->OsdTerm);
         logPrintf (3, "info cleared\n");
      break;

      case COMCODE_OPEN_DEPENDENT_DEVICES:
         OpenDependentInputDevices (Command->Value);
         logPrintf (3, "dependent devices of %s reopened\n",
                      xName (OutputDevice));
      break;

      case COMCODE_OPEN_VIDEOTEXT:
         FindCommand (NewCommand, COMCODE_ENTER_VIDEOTEXT);
         if (!NewCommand)
         {
            if (OutputDevice->Channel->VtxPid)
            {
               sprintf (sbuf, " \f%c%s\f%c\n",
                  COL2ID (clrCyan), screenMsgs[OPEN_VTX]
                  [OutputDevice->osdLanguage], COL2ID (clrWhite));
               ScheduleDirectText (COMCODE_INFO_MESSAGE, sbuf);
               PurgeCommands (COMCODE_CHANNEL_SET);
               OpenDependentInputDevices (IMPL_VTX);
               ScheduleDelayed (COMCODE_ENTER_VIDEOTEXT, 10);
            } else {
               sprintf (sbuf, " \f%c%s\f%c\n",
                  COL2ID (clrRed), screenMsgs[CANT_OPEN_VTX]
                  [OutputDevice->osdLanguage], COL2ID (clrWhite));
               ScheduleDirectText (COMCODE_INFO_MESSAGE, sbuf);
            }
         }
      break;

      case COMCODE_OPEN_EIT:
         OpenDependentInputDevices (IMPL_EIT);
      break;

      case COMCODE_ENTER_CHANNEL_MENU:
         OutputDevice->InfoChannel = OutputDevice->Channel;
         PurgeCommands (COMCODE_PROGRAM_INFO);
         PurgeCommands (COMCODE_CLEAR_INFO);
         PurgeCommands (COMCODE_CHANNEL_SET);
         ScheduleDirect (COMCODE_SHOW_CHANNEL_MENU);
         OutputDevice->State = OUTST_CHANNEL_MENU;
      break;

      case COMCODE_SHOW_CHANNEL_MENU:
         if (!(sptr = getChannelMenu (OutputDevice->InfoChannel))) break;
         osdTermClear (OutputDevice->OsdTerm);
         if (!(tptr = getExtendedInfo (OutputDevice->InfoChannel)))
            osdTermMenuOut (OutputDevice->OsdTerm, sptr);
         else
         {
            sprintf (sbuf, "%s\n%s", sptr, tptr);
            osdTermMenuOut (OutputDevice->OsdTerm, sbuf);
         }
         osdClearEow (OutputDevice);
         osdShow (OutputDevice->osdDev);              
         gettimeofday (&(OutputDevice->TimeStamp), NULL);
      break;

      case COMCODE_CHANNEL_MENU_DOWN:
         Channel = OutputDevice->InfoChannel;
         if (Channel && xPred (Channel))
         {
            OutputDevice->InfoChannel =
               (struct Channel *) xPred (Channel);
            ScheduleDirect (COMCODE_SHOW_CHANNEL_MENU);
         }
      break;

      case COMCODE_CHANNEL_MENU_UP:
         Channel = OutputDevice->InfoChannel;
         if (Channel && xSucc (Channel))
         {
            OutputDevice->InfoChannel =
               (struct Channel *) xSucc (Channel);
            ScheduleDirect (COMCODE_SHOW_CHANNEL_MENU);
         }
      break;

      case COMCODE_QUIT_CHANNEL_MENU:
         osdHide (OutputDevice->osdDev);
         osdClear (OutputDevice->osdDev);
         osdTermClear (OutputDevice->OsdTerm);
         OutputDevice->InfoChannel = NULL;
         OutputDevice->State = OUTST_IDLE;
      break;

      case COMCODE_USE_CHANNEL_MENU:
         if (OutputDevice->InfoChannel)
         {
            ScheduleDirectValue (COMCODE_CHANNEL_INFO,
               OutputDevice->InfoChannel->ChannelNumber);
            ScheduleDirectValue (COMCODE_CHANNEL_SET,
               OutputDevice->InfoChannel->ChannelNumber);
         }
         osdHide (OutputDevice->osdDev);
         osdClear (OutputDevice->osdDev);
         osdTermClear (OutputDevice->OsdTerm);
         OutputDevice->InfoChannel = NULL;
         OutputDevice->State = OUTST_IDLE;
      break;

      case COMCODE_ENTER_LANGUAGE_MENU:
         if (OutputDevice->Channel->LanguageList)
         {
            OutputDevice->InfoChannel = (struct Channel *)
               OutputDevice->Channel->LanguageList->Head;
            PurgeCommands (COMCODE_PROGRAM_INFO);
            PurgeCommands (COMCODE_CLEAR_INFO);
            PurgeCommands (COMCODE_CHANNEL_SET);
            ScheduleDirect (COMCODE_SHOW_LANGUAGE_MENU);
            OutputDevice->State = OUTST_LANGUAGE_MENU;
         }
      break;

      case COMCODE_SHOW_LANGUAGE_MENU:
         osdTermMenuOut (OutputDevice->OsdTerm,
            getChannelMenu (OutputDevice->InfoChannel));              
         osdClearEow (OutputDevice);
         osdShow (OutputDevice->osdDev);              
         gettimeofday (&(OutputDevice->TimeStamp), NULL);
      break;

      case COMCODE_LANGUAGE_MENU_DOWN:
         Channel = OutputDevice->InfoChannel;
         if (Channel && xPred (Channel))
         {
            OutputDevice->InfoChannel =
               (struct Channel *) xPred (Channel);
            ScheduleDirect (COMCODE_SHOW_LANGUAGE_MENU);
         }
      break;

      case COMCODE_LANGUAGE_MENU_UP:
         Channel = OutputDevice->InfoChannel;
         if (Channel && xSucc (Channel))
         {
            OutputDevice->InfoChannel =
               (struct Channel *) xSucc (Channel);
            ScheduleDirect (COMCODE_SHOW_LANGUAGE_MENU);
         }
      break;

      case COMCODE_QUIT_LANGUAGE_MENU:
         osdHide (OutputDevice->osdDev);
         osdClear (OutputDevice->osdDev);
         osdTermClear (OutputDevice->OsdTerm);
         OutputDevice->InfoChannel = NULL;
         OutputDevice->State = OUTST_IDLE;
      break;

      case COMCODE_USE_LANGUAGE_MENU:
         if (OutputDevice->InfoChannel)
         {
            OutputDevice->Channel->AudioPid =
               OutputDevice->InfoChannel->AudioPid;
            ScheduleDirectValue (COMCODE_CHANNEL_SET,
               OutputDevice->Channel->ChannelNumber);
         }
         osdHide (OutputDevice->osdDev);
         osdClear (OutputDevice->osdDev);
         osdTermClear (OutputDevice->OsdTerm);
         OutputDevice->InfoChannel = NULL;
         OutputDevice->State = OUTST_IDLE;
      break;

      case COMCODE_ENTER_NVOD_MENU:
         if (OutputDevice->Channel->NvodList)
         {
            OutputDevice->InfoChannel = (struct Channel *)
               OutputDevice->Channel->NvodList->Head;
            PurgeCommands (COMCODE_PROGRAM_INFO);
            PurgeCommands (COMCODE_CLEAR_INFO);
            PurgeCommands (COMCODE_CHANNEL_SET);
            ScheduleDirect (COMCODE_SHOW_NVOD_MENU);
            OutputDevice->State = OUTST_NVOD_MENU;
         }
      break;

      case COMCODE_SHOW_NVOD_MENU:
         if (!(sptr = getChannelMenu (OutputDevice->InfoChannel))) break;
         osdTermClear (OutputDevice->OsdTerm);
         if (!(tptr = getExtendedInfo (OutputDevice->Channel)))
            osdTermMenuOut (OutputDevice->OsdTerm, sptr);
         else
         {
            sprintf (sbuf, "%s\n%s", sptr, tptr);
            osdTermMenuOut (OutputDevice->OsdTerm, sbuf);
         }
         osdClearEow (OutputDevice);
         osdShow (OutputDevice->osdDev);              
         gettimeofday (&(OutputDevice->TimeStamp), NULL);
      break;

      case COMCODE_NVOD_MENU_DOWN:
         Channel = OutputDevice->InfoChannel;
         if (Channel && xPred (Channel))
         {
            OutputDevice->InfoChannel =
               (struct Channel *) xPred (Channel);
            ScheduleDirect (COMCODE_SHOW_NVOD_MENU);
         }
      break;

      case COMCODE_NVOD_MENU_UP:
         Channel = OutputDevice->InfoChannel;
         if (Channel && xSucc (Channel))
         {
            OutputDevice->InfoChannel =
               (struct Channel *) xSucc (Channel);
            ScheduleDirect (COMCODE_SHOW_NVOD_MENU);
         }
      break;

      case COMCODE_QUIT_NVOD_MENU:
         osdHide (OutputDevice->osdDev);
         osdClear (OutputDevice->osdDev);
         osdTermClear (OutputDevice->OsdTerm);
         OutputDevice->InfoChannel = NULL;
         OutputDevice->State = OUTST_IDLE;
      break;

      case COMCODE_USE_NVOD_MENU:
         if (OutputDevice->InfoChannel)
         {
            OutputDevice->Channel->TransportStreamID =
               OutputDevice->InfoChannel->TransportStreamID;
            OutputDevice->Channel->Frequency =
               OutputDevice->InfoChannel->Frequency;
            OutputDevice->Channel->SymbolRate =
               OutputDevice->InfoChannel->SymbolRate;
            OutputDevice->Channel->Polarisation =
               OutputDevice->InfoChannel->Polarisation;
            OutputDevice->Channel->ServiceID =
               OutputDevice->InfoChannel->ServiceID;
            OutputDevice->Channel->AudioPid =
               OutputDevice->InfoChannel->AudioPid;
            OutputDevice->Channel->VideoPid =
               OutputDevice->InfoChannel->VideoPid;
            ScheduleDirectValue (COMCODE_CHANNEL_INFO,
               OutputDevice->Channel->ChannelNumber);
            ScheduleDirectValue (COMCODE_CHANNEL_SET,
               OutputDevice->Channel->ChannelNumber);
         }
         osdHide (OutputDevice->osdDev);
         osdClear (OutputDevice->osdDev);
         osdTermClear (OutputDevice->OsdTerm);
         OutputDevice->InfoChannel = NULL;
         OutputDevice->State = OUTST_IDLE;
      break;

      case COMCODE_ENTER_LINK_MENU:
         if (OutputDevice->Channel->LinkList)
         {
            OutputDevice->InfoChannel = (struct Channel *)
               OutputDevice->Channel->LinkList->Head;
            PurgeCommands (COMCODE_PROGRAM_INFO);
            PurgeCommands (COMCODE_CLEAR_INFO);
            PurgeCommands (COMCODE_CHANNEL_SET);
            ScheduleDirect (COMCODE_SHOW_NVOD_MENU);
            OutputDevice->State = OUTST_NVOD_MENU;
         }
      break;

      case COMCODE_ENTER_VIDEOTEXT:
         if (OutputDevice->VideoText)
         {
            nextPage = 0x10000;

            PurgeCommands (COMCODE_PROGRAM_INFO);
            PurgeCommands (COMCODE_CLEAR_INFO);
            osdClear (OutputDevice->osdDev);
            osdShow (OutputDevice->osdDev);              
            if (VtxPage = OutputDevice->VideoText->DisplayedPage)
             { nextPage = (VtxPage->PageNumber & 0xFFF) << 8; }
            OutputDevice->VideoText->DisplayedPage = NULL;
            ScheduleDirectValue (COMCODE_SHOW_VIDEOTEXT, nextPage);
            logPrintf (3, "Videotext state entered (%d)\n", nextPage);
            OutputDevice->State = OUTST_VIDEOTEXT;
         }
         else CloseDependentInputDevices (IMPL_VTX);
      break;

      case COMCODE_SHOW_VIDEOTEXT:
         if (OutputDevice->VideoText)
         {
            curPage = (Command->Value >> 8) & 0xFFF;
            curSub = Command->Value & 0xFF;
            nextPage = Command->Value;

            OutputDevice->VideoText->PageNumber = curPage;
            xForeach (OutputDevice->VideoText->Pages, VtxPage)
               if (((VtxPage->PageNumber & 0xFFF) == curPage) &&
                   (VtxPage->Flags & PG_READY) && (!curSub ||
                   ((VtxPage->SubpageNumber & 0xFF) == curSub))) break;
            if (VtxPage && ((VtxPage != OutputDevice->VideoText->DisplayedPage)
                 || (VtxPage->Flags & PG_CHANGED)))
            {
               sptr = prepare_osd (OutputDevice->VideoText, VtxPage);
               osdVideoTextOut (OutputDevice, sptr);
               OutputDevice->VideoText->DisplayedPage = VtxPage;
               VtxPage->Flags &= ~(PG_CHANGED);
            }
            else
            {
               sptr = prepare_osd_header (OutputDevice->VideoText, VtxPage);
               osdVideoTextHeaderOut (OutputDevice, sptr);
            }
            xMemFree (sptr);
            ScheduleDelayedValue (COMCODE_SHOW_VIDEOTEXT, nextPage, 2);
         }
      break;

      case COMCODE_VTX_PAGE_UP:
         if (OutputDevice->VideoText)
         {
            struct VtxPage *VtxPage;

            if ((VtxPage = OutputDevice->VideoText->DisplayedPage) &&
                (VtxPage = (struct VtxPage *) xSucc (VtxPage)))
            {
               PurgeCommands (COMCODE_SHOW_VIDEOTEXT);
               ScheduleDirectValue (COMCODE_SHOW_VIDEOTEXT,
                  ((VtxPage->PageNumber & 0xFFF) << 8));
            }
         }
      break;

      case COMCODE_VTX_PAGE_DOWN:
         if (OutputDevice->VideoText)
         {
            struct VtxPage *VtxPage;

            if ((VtxPage = OutputDevice->VideoText->DisplayedPage) &&
                (VtxPage = (struct VtxPage *) xPred (VtxPage)))
            {
               PurgeCommands (COMCODE_SHOW_VIDEOTEXT);
               ScheduleDirectValue (COMCODE_SHOW_VIDEOTEXT,
                  ((VtxPage->PageNumber & 0xFFF) << 8));
            }
         }
      break;

      case COMCODE_VTX_SET_PAGE:
      break;

      case COMCODE_QUIT_VIDEOTEXT:
         PurgeCommands (COMCODE_SHOW_VIDEOTEXT);
         osdHide (OutputDevice->osdDev);
         osdClear (OutputDevice->osdDev);
         OutputDevice->State = OUTST_IDLE;
         CloseDependentInputDevices (IMPL_VTX);
      break;

      case COMCODE_ENTER_USER_MENU:
         if (OutputDevice->UserMenu->Head &&
             (((struct MenuItem *)OutputDevice->UserMenu->Head)
                ->Type == MTYP_MENU) &&
             (OutputDevice->ActualMenuItem =
             (struct MenuItem *)((struct MenuItem *)
             OutputDevice->UserMenu->Head)->Menu->Head))
         {
            PurgeCommands (COMCODE_PROGRAM_INFO);
            PurgeCommands (COMCODE_CLEAR_INFO);
            PurgeCommands (COMCODE_CHANNEL_SET);
            ScheduleDirect (COMCODE_SHOW_USER_MENU);
            OutputDevice->State = OUTST_USER_MENU;
         }
      break;

      case COMCODE_SHOW_USER_MENU:
         if (!(sptr = getUserMenu (OutputDevice))) break;
         osdTermClear (OutputDevice->OsdTerm);
         osdTermMenuOut (OutputDevice->OsdTerm, sptr);
         osdClearEow (OutputDevice);
         osdShow (OutputDevice->osdDev);              
         gettimeofday (&(OutputDevice->TimeStamp), NULL);
      break;

      case COMCODE_USER_MENU_DOWN:
         if (OutputDevice->ActualMenuItem &&
            xSucc (OutputDevice->ActualMenuItem))
         {
            OutputDevice->ActualMenuItem =
               (struct MenuItem *) xSucc (OutputDevice->ActualMenuItem);
            ScheduleDirect (COMCODE_SHOW_USER_MENU);
         }
      break;

      case COMCODE_USER_MENU_UP:
         if (OutputDevice->ActualMenuItem &&
            xPred (OutputDevice->ActualMenuItem))
         {
            OutputDevice->ActualMenuItem =
               (struct MenuItem *) xPred (OutputDevice->ActualMenuItem);
            ScheduleDirect (COMCODE_SHOW_USER_MENU);
         }
      break;

      case COMCODE_QUIT_USER_MENU:
         if (OutputDevice->ActualMenuItem &&
             OutputDevice->ActualMenuItem->UplevelMenu &&
             OutputDevice->ActualMenuItem->UplevelMenu->UplevelMenu)
         {
            OutputDevice->ActualMenuItem = 
               OutputDevice->ActualMenuItem->UplevelMenu;
            ScheduleDirect (COMCODE_SHOW_USER_MENU);
            break;
         }
         osdHide (OutputDevice->osdDev);
         osdClear (OutputDevice->osdDev);
         osdTermClear (OutputDevice->OsdTerm);
         OutputDevice->ActualMenuItem = NULL;
         OutputDevice->State = OUTST_IDLE;
      break;

      case COMCODE_USE_USER_MENU:
         if (OutputDevice->ActualMenuItem &&
               (OutputDevice->ActualMenuItem->Type == MTYP_MENU) &&
             (OutputDevice->ActualMenuItem = (struct MenuItem *)
                 OutputDevice->ActualMenuItem->Menu->Head))
         {
            ScheduleDirect (COMCODE_SHOW_USER_MENU);
            break;
         }
         osdHide (OutputDevice->osdDev);
         osdClear (OutputDevice->osdDev);
         osdTermClear (OutputDevice->OsdTerm);
         if (OutputDevice->ActualMenuItem &&
             OutputDevice->ActualMenuItem->Command)
         {
            Dcl_Eval (interp, OutputDevice->ActualMenuItem->Command);
            if (interp->result && interp->result[0])
            {
               sprintf (sbuf, "\f%c%s\n", COL2ID (clrWhite), interp->result);
               ScheduleDirectText (COMCODE_INFO_TEXT, sbuf);
            }
         }
         OutputDevice->ActualMenuItem = NULL;
         OutputDevice->State = OUTST_IDLE;
      break;
   }

#undef ImplicitOutputID

   CurrentOutputDevice = NULL;

   xRemove (CommandQueue, Command);
   if (Command->Code == COMCODE_QUIT_PROGRAM) return (1);
   return (0);
}

