//////////////////////////////////////////////////////////////
///                                                        ///
/// dvb_si.c: routines to parse the DVB-SI stream          ///
///                                                        ///
//////////////////////////////////////////////////////////////

// $Revision: 1.4 $
// $Date: 2001/06/16 14:20:08 $
// $Author: hakenes $
//
//   (C) 2001 Rolf Hakenes <hakenes@hippomi.de>, under the GNU GPL.
//
// dtv is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// dtv is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You may have received a copy of the GNU General Public License
// along with dtv; see the file COPYING.  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#include <ost/dmx.h>

#include <libsi.h>
#include <si_tables.h>
#include "dvb_si.h"

#include "definitions.h"
#include "structures.h"
#include "prototypes.h"



static u_char siBuffer[MAX_SECTION_BUFFER];  /* maximum section */


int openTdt(struct InputDevice *InputDevice)
{
   SetupFilter (InputDevice, PID_TDT, TID_TDT);

   return (InputDevice->FileHandle);
}


int openPfo(struct InputDevice *InputDevice)
{
   struct OutputDevice                   *OutputDevice;

   FindOutputDevice (InputDevice, OutputDevice);

   if (!OutputDevice || !OutputDevice->Channel)
      return (-1);

   SetupFilterAndSection (InputDevice, PID_SDT, TID_SDT_ACT, 0);

   InputDevice->State = INPDEV_GET_SDT;
   if (InputDevice->RememberKey) 
   {
      xMemFreeAll (&InputDevice->RememberKey);
      InputDevice->ProgramList = NULL;
      InputDevice->Program = NULL;
      InputDevice->ServiceList = NULL;
      InputDevice->Service = NULL;
      InputDevice->EventList = NULL;
      InputDevice->Event = NULL;
   }

   return (InputDevice->FileHandle);
}


int openEit(struct InputDevice *InputDevice)
{
   struct OutputDevice                   *OutputDevice;
   struct Channel                        *GlobChannel;
   struct ProgramInfo                    *ProgramInfo, *NextProgInfo;
   struct MEM_CHUNK                     **OriginalKey;
   struct timeval                         CurrentTime;
        
   FindOutputDevice (InputDevice, OutputDevice);

   if (!OutputDevice || !OutputDevice->Channel ||
       !(GlobChannel = OutputDevice->Channel->MasterChannel))
      return (-1);

   gettimeofday (&CurrentTime, NULL);
   OriginalKey = xGetRemember ();
   xSetRemember (&GlobChannel->ProgramRemember);
   if (GlobChannel->ProgramInfos &&
       (ProgramInfo = (struct ProgramInfo *) GlobChannel->ProgramInfos->Head))
   do
   {
      NextProgInfo = (struct ProgramInfo *) xSucc (ProgramInfo);
      if ((ProgramInfo->StartTime + ProgramInfo->Duration) < CurrentTime.tv_sec)
      {
         if (ProgramInfo->Content) xMemFree (ProgramInfo->Content);
         if (ProgramInfo->ShortName) xMemFree (ProgramInfo->ShortName);
         if (ProgramInfo->ShortText) xMemFree (ProgramInfo->ShortText);
         if (ProgramInfo->ExtendedName) xMemFree (ProgramInfo->ExtendedName);
         if (ProgramInfo->ExtendedText) xMemFree (ProgramInfo->ExtendedText);
         xRemove (GlobChannel->ProgramInfos, ProgramInfo);
         xMemFree (ProgramInfo);
      }
   } while (ProgramInfo = NextProgInfo);
   xSetRemember (OriginalKey);

   SetupFilterOnEitID (InputDevice, 0, GlobChannel->ServiceID);

   InputDevice->UserData = GetGlobalOption ("eitAcquisitionNumberOfFails");
   InputDevice->State = INPDEV_GET_EIT_ACT;
   if (InputDevice->RememberKey) 
   {
      xMemFreeAll (&InputDevice->RememberKey);
      InputDevice->ProgramList = NULL;
      InputDevice->Program = NULL;
      InputDevice->ServiceList = NULL;
      InputDevice->Service = NULL;
      InputDevice->EventList = NULL;
      InputDevice->Event = NULL;
   }
   return (InputDevice->FileHandle);
}


void closeSi(struct InputDevice *InputDevice)
{
   if (InputDevice->FileHandle)
   {
      ioctl (InputDevice->FileHandle, DMX_STOP, 0);
      close (InputDevice->FileHandle);
      InputDevice->FileHandle = 0;
      InputDevice->Status = INPDEV_DISABLED;
   }
   return;
}


int getTdtEvent (struct InputDevice *InputDevice)
{
   struct timeval         TempTime, CurrentTime;
   time_t                 NewTime;
   int                    count, DiffTime;
   struct Command        *Command;
        
   if ((count = read (InputDevice->FileHandle, siBuffer, MAX_SECTION_BUFFER)) < 0)
      return (0);

   if (NewTime = siParseTDT (siBuffer))
   {
      gettimeofday (&CurrentTime, NULL);

      TempTime.tv_sec = NewTime;
      TempTime.tv_usec = 0;

      settimeofday (&TempTime, NULL);
      logPrintf (5, "Time adjusted to %s", ctime (&NewTime));
      closeSi (InputDevice);

      DiffTime = CurrentTime.tv_sec - NewTime;
      xForeach (CommandQueue, Command)
         Command->TimeStamp.tv_sec -= DiffTime;
   }

   return (0);
}


int getPfoEvent (struct InputDevice *InputDevice)
{
   struct OutputDevice                   *OutputDevice;
   struct LIST                           *Result;
   struct MEM_CHUNK                     **OriginalKey;
   unsigned short                         flag, EventListSize;
   int                                    count;
        
   if ((count = read (InputDevice->FileHandle, siBuffer, MAX_SECTION_BUFFER)) < 0)
      return (0);

#define ImplicitOutputID OutputDevice->Identifier

   FindOutputDevice (InputDevice, OutputDevice);

   if (OutputDevice)
   {
      OriginalKey = xGetRemember ();
      xSetRemember (&InputDevice->RememberKey);
      switch (InputDevice->State)
      {
         case INPDEV_GET_SDT:
         logPrintf (4, "InpDev Pfo GetSDT\n");
         {
            struct Service           *Service;

            if (Result = siParseSDT (siBuffer))
               xForeach (Result, Service)
               {
                  if (Service->ServiceID !=
                         OutputDevice->Channel->ServiceID) continue;

                  logPrintf (6, "Service %d identified\n",
                     Service->ServiceID);
                  InputDevice->Service = Service;
                  break;
               }
            xConCat (InputDevice->ServiceList, Result);
            if (Result && TableHasMoreSections (siBuffer))
               SetupFilterAndSection (InputDevice, PID_SDT,
                   TID_SDT_ACT, GetSectionNumber(siBuffer)+1);
            else
            {
               SetupFilterAndSection (InputDevice, PID_PAT, TID_PAT, 0);
               InputDevice->State = INPDEV_GET_PAT;
            }
         }
         break;

         case INPDEV_GET_PAT:
         logPrintf (4, "InpDev Pfo GetPAT\n");
         {
            struct Program           *Program;

            if (Result = siParsePAT (siBuffer))
               xForeach (Result, Program)
               {
                  if (Program->ProgramID !=
                         OutputDevice->Channel->ServiceID) continue;

                  logPrintf (6, "Program %d identified\n",
                         Program->ProgramID);

                  InputDevice->Program = Program;
                  break;
               }
            xConCat (InputDevice->ProgramList, Result);
            if (Result && TableHasMoreSections (siBuffer))
               SetupFilterAndSection (InputDevice, PID_PAT,
                   TID_PAT, GetSectionNumber(siBuffer)+1);
            else
            {
               struct Descriptor *Descriptor;

               if (!InputDevice->Service)
               {
                  if (Program = InputDevice->Program)
                  {
                     CreateService (InputDevice->Service,
                                       Program->ProgramID,
                                       Program->TransportStreamID,
                                       0, 0, 0);
                     OutputDevice->Channel->TransportStreamID =
                        Program->TransportStreamID;
                     SetupFilterOnID (InputDevice, Program->NetworkPID,
                        TID_PMT, 0, Program->ProgramID);
                     InputDevice->State = INPDEV_GET_TV_PMT;
                  }
                  else
                  {
                     logPrintf (2, "Service %d not found\n",
                        OutputDevice->Channel->ServiceID);
                     InputDevice->State = INPDEV_ERROR;
                     closeSi (InputDevice);
                  }
                  break;
               }
               OutputDevice->Channel->TransportStreamID =
                  InputDevice->Service->TransportStreamID;
               if (GetRunningStatus (InputDevice->Service->Status)
                     != RUNNING_STATUS_RUNNING)
               {
                  logPrintf (3, "Service %d not running\n",
                     InputDevice->Service->ServiceID);
                  sprintf (siBuffer, " \f%c%s\f%c\n",
                     COL2ID (clrRed), screenMsgs[SERVICE_NOT_ACTIVE]
                     [OutputDevice->osdLanguage], COL2ID (clrWhite));
                  ScheduleDirectText (COMCODE_INFO_MESSAGE, siBuffer);
                  InputDevice->State = INPDEV_FINISHED;
                  ScheduleDirect (COMCODE_OPEN_EIT);
                  closeSi (InputDevice);
                  break;
               }
               xForeach (InputDevice->Service->Descriptors, Descriptor)
                  if (Descriptor->Tag == DESCR_SERVICE) break;
               if (!Descriptor)
               {
                  xForeach (InputDevice->Service->Descriptors, Descriptor)
                     if (Descriptor->Tag == DESCR_TIME_SHIFTED_SERVICE)
                        break;
                  if (!Descriptor)
                  {
                     logPrintf (2, "Service %d has no descriptor\n",
                        InputDevice->Service->ServiceID);
                     InputDevice->State = INPDEV_ERROR;
                     closeSi (InputDevice);
                     break;
                  }
                  if (Program = InputDevice->Program)
                  {
                     SetupFilterOnID (InputDevice, Program->NetworkPID,
                        TID_PMT, 0, Program->ProgramID);
                     InputDevice->State = INPDEV_GET_TV_PMT;
                  }
                  else
                  {
                     logPrintf (2, "Service %d has no PMT entry\n",
                        InputDevice->Service->ServiceID);
                     InputDevice->State = INPDEV_ERROR;
                     closeSi (InputDevice);
                  }
                  break;
               }
               switch (((struct ServiceDescriptor*)Descriptor)->ServiceType)
               {
                  case 0x01:    /* TV */
                     if (Program = InputDevice->Program)
                     {
                        SetupFilterOnID (InputDevice, Program->NetworkPID,
                           TID_PMT, 0, Program->ProgramID);
                        InputDevice->State = INPDEV_GET_TV_PMT;
                     }
                     else
                     {
                        logPrintf (2, "Service %d has no PMT entry\n",
                           InputDevice->Service->ServiceID);
                        InputDevice->State = INPDEV_ERROR;
                        closeSi (InputDevice);
                     }
                  break;

                  case 0x04:    /* NVOD */
                  {
                     struct NvodReferenceDescriptor *NvodReferenceDescr;
                     struct NvodReferenceItem       *Item;
                     struct Channel                 *chan;

                     xForeach (InputDevice->Service->Descriptors,
                                  Descriptor)
                        if (Descriptor->Tag == DESCR_NVOD_REF) break;
                     if (!Descriptor)
                     {
                        logPrintf (2, "NVOD Service %d has no Channels\n",
                           InputDevice->Service->ServiceID);
                        InputDevice->State = INPDEV_ERROR;
                        ScheduleDirect (COMCODE_OPEN_EIT);
                        closeSi (InputDevice);
                        break;
                     }
                     NvodReferenceDescr = 
                        (struct NvodReferenceDescriptor *) Descriptor;
                     xSetRemember (&OutputDevice->Channel->NvodRemember);
                     if (OutputDevice->Channel->NvodRemember) 
                       xMemFreeAll (&OutputDevice->Channel->NvodRemember);
                     OutputDevice->Channel->NvodList = xNewList (NULL);
                     xForeach (NvodReferenceDescr->Items, Item)
                        if (chan = getChannelFromNvodItem (Item,
                              OutputDevice->Channel))
                        {
                           xSetName (chan, screenMsgs[TO_BE_SCANNED]
                              [OutputDevice->osdLanguage]);
                           xAddTail (OutputDevice->Channel->NvodList,
                                        chan);
                        }
                     if (!xSize (OutputDevice->Channel->NvodList))
                     {
                        OutputDevice->Channel->NvodList = NULL;
                        logPrintf (2, "NVOD Service %d has no Channels\n",
                           InputDevice->Service->ServiceID);
                        InputDevice->State = INPDEV_ERROR;
                        ScheduleDirect (COMCODE_OPEN_EIT);
                        closeSi (InputDevice);
                     }
                     else
                     {
                        CloneChannel (chan, OutputDevice->Channel,
                           OutputDevice->Channel->VideoPid, 0,
                           OutputDevice->Channel->AudioPid, 0,
                           screenMsgs[REPEATED_SCAN]
                           [OutputDevice->osdLanguage]);
                        xAddTail (OutputDevice->Channel->NvodList, chan);
                        logPrintf (6, "Service %d is NVOD\n",
                           InputDevice->Service->ServiceID);
                        sprintf (siBuffer, " \f%c%s\f%c\n",
                           COL2ID (clrGreen), screenMsgs[VOD_AVAILABLE]
                           [OutputDevice->osdLanguage], COL2ID(clrWhite));
                        ScheduleDirectText (COMCODE_INFO_MESSAGE, siBuffer);
                        SetupFilterOnEitID (InputDevice, 
                            0, InputDevice->Service->ServiceID);
                        InputDevice->State = INPDEV_GET_TV_EIT;
                     }
                  }
                  break;

                  case 0x02:    /* Radio */
                  default:
                     logPrintf (6, "Service %d not TV (skipped)\n",
                        InputDevice->Service->ServiceID);
                     sprintf (siBuffer, " \f%c%s\f%c\n",
                        COL2ID(clrGreen), screenMsgs[SERVICE_IS_RADIO]
                        [OutputDevice->osdLanguage], COL2ID(clrWhite));
                     ScheduleDirectText (COMCODE_INFO_MESSAGE, siBuffer);
                     InputDevice->State = INPDEV_FINISHED;
                     ScheduleDirect (COMCODE_OPEN_EIT);
                     closeSi (InputDevice);
                  break;
               }
            }
         }
         break;

         case INPDEV_GET_TV_PMT:
         logPrintf (4, "InpDev Pfo GetTVPMT\n");
         {
            struct Pid                 *Pid;
            struct PidInfo             *PidInfo;
            struct Program             *Program;
            struct Service             *Service;
            struct Descriptor          *Descriptor;
            struct Channel             *chan;
            int                         i, j;
            int                         VideoPids = 0, AudioPids = 0;
            int                         VideoPid[MAX_PIDS];
            int                         AudioPid[MAX_PIDS];
            int                         VideoTag[MAX_PIDS];
            int                         AudioTag[MAX_PIDS];
            char                       *Language[MAX_PIDS];

            if (!(Program = InputDevice->Program) ||
                !(Service = InputDevice->Service) ||
                !(Pid = siParsePMT (siBuffer)))
             { logPrintf (2, "corrupt PMT entry\n");
               InputDevice->State = INPDEV_ERROR;
               closeSi (InputDevice); break; }

            xAddTail (Program->Pids, Pid);

            flag = 0;
            if (Pid->ProgramID == InputDevice->Service->ServiceID)
               xForeach (Pid->InfoList, PidInfo)
                  switch (PidInfo->StreamType)
                  {
                     case STREAMTYPE_ISO_VIDEO:
                     case STREAMTYPE_13818_VIDEO:
                     case STREAMTYPE_13522_MPEG:
                        xForeach (PidInfo->Descriptors, Descriptor)
                        {
                           switch (Descriptor->Tag)
                           {
                              case DESCR_STREAM_ID:
                                 VideoTag[VideoPids] = ((struct
                                    StreamIdentifierDescriptor *)
                                    Descriptor)->ComponentTag;
                              break;
                           }
                        }
                        VideoPid[VideoPids++] = PidInfo->ElementaryPid;
                     break;

                     case STREAMTYPE_11172_AUDIO:
                     case STREAMTYPE_13818_AUDIO:
                        xForeach (PidInfo->Descriptors, Descriptor)
                        {
                           switch (Descriptor->Tag)
                           {
                              case DESCR_ISO_639_LANGUAGE:
                                 Language[AudioPids] = ((struct
                                    Iso639LanguageDescriptor *)
                                    Descriptor)->LanguageCode;
                              break;

                              case DESCR_STREAM_ID:
                                 AudioTag[AudioPids] = ((struct
                                    StreamIdentifierDescriptor *)
                                    Descriptor)->ComponentTag;
                              break;
                           }
                        }
                        AudioPid[AudioPids++] = PidInfo->ElementaryPid;
                     break;

                     case STREAMTYPE_VIDEOTEXT:
                        if (!OutputDevice->Channel->VtxPid ||
                             OutputDevice->Channel->VtxPid != 
                                PidInfo->ElementaryPid)
                        {
                           OutputDevice->Channel->VtxPid =
                              PidInfo->ElementaryPid;
                           logPrintf (4, "Corrected VtxPID to %d\n",
                              PidInfo->ElementaryPid);
                        }
                     case STREAMTYPE_ITU_222:
                     case STREAMTYPE_13818_A:
                     case STREAMTYPE_13818_B:
                     case STREAMTYPE_13818_C:
                     case STREAMTYPE_13818_D:
                     case STREAMTYPE_13818_AUX:
                     break;
                  }
            if (!VideoPids || !AudioPids)
            {
               logPrintf (2, "Service %d has no Video/AudioPID pair\n",
                             Service->ServiceID);
               InputDevice->State = INPDEV_ERROR;
               closeSi (InputDevice);
               break;
            }
            flag = 0;
            for (i = 0; i < VideoPids; i++)
            {
               logPrintf (6, "Service %d Pid: Video %d\n",
                            Service->ServiceID, VideoPid[i]);
               if (VideoPid[i] == OutputDevice->Channel->VideoPid)
                  flag |= 1;
            }
            for (i = 0; i < AudioPids; i++)
            {
               logPrintf (6, "Service %d Pid: Audio %d\n",
                            Service->ServiceID, AudioPid[i]);
               if (AudioPid[i] == OutputDevice->Channel->AudioPid)
                  flag |= 2;
            }
            if (!(flag&1)) OutputDevice->Channel->VideoPid = VideoPid[0];
            if (!(flag&2)) OutputDevice->Channel->AudioPid = AudioPid[0];
            if (flag != 3)
            {
               logPrintf (3, "Service %d adapted: Video %d Audio %d\n",
                            Service->ServiceID, VideoPid[0], AudioPid[0]);
               ScheduleDirectValue (COMCODE_CHANNEL_SET,
                     OutputDevice->Channel->ChannelNumber);
               InputDevice->State = INPDEV_FINISHED;
               closeSi (InputDevice);
            }
            else
            {
               if ((VideoPids > 1) || (AudioPids > 1))
               {
                  xSetRemember (&OutputDevice->Channel->LanguageRemember);
                  if (OutputDevice->Channel->LanguageRemember)
                   xMemFreeAll (&OutputDevice->Channel->LanguageRemember);
                  OutputDevice->Channel->LanguageList = xNewList (NULL);
                  for (i = 0; i < VideoPids; i++)
                     for (j = 0; j < AudioPids; j++)
                     {
                        chan = NULL;
                        if (!strcasecmp (Language[j], "deu") ||
                            !strcasecmp (Language[j], "ger")) {
                           CloneChannel (chan, OutputDevice->Channel,
                              VideoPid[i], VideoTag[i], AudioPid[j],
                              AudioTag[j], "Deutsch");
                        }
                        if (!strcasecmp (Language[j], "eng")) {
                           CloneChannel (chan, OutputDevice->Channel,
                              VideoPid[i], VideoTag[i], AudioPid[j],
                              AudioTag[j], "Englisch");
                        } 
                        if (!strcasecmp (Language[j], "fra") ||
                            !strcasecmp (Language[j], "fre")) {
                           CloneChannel (chan, OutputDevice->Channel,
                              VideoPid[i], VideoTag[i], AudioPid[j],
                              AudioTag[j], "Franzsisch");
                        }
                        if (!strcasecmp (Language[j], "esl") ||
                            !strcasecmp (Language[j], "spa")) {
                           CloneChannel (chan, OutputDevice->Channel,
                              VideoPid[i], VideoTag[i], AudioPid[j],
                              AudioTag[j], "Spanisch");
                        }
                        if (!strcasecmp (Language[j], "dut")) {
                           CloneChannel (chan, OutputDevice->Channel,
                              VideoPid[i], VideoTag[i], AudioPid[j],
                              AudioTag[j], "Niederlndisch");
                        }
                        if (!chan) {
                           CloneChannel (chan, OutputDevice->Channel,
                              VideoPid[i], VideoTag[i], AudioPid[j],
                              AudioTag[j], Language[j]);
                        }
                        xAddTail (OutputDevice->Channel->LanguageList,
                           chan);
                     }

                  sprintf (siBuffer, " \f%c%s\f%c\n",
                     COL2ID (clrYellow), screenMsgs[MULTIPLE_AUDIO_PIDS]
                     [OutputDevice->osdLanguage], COL2ID (clrWhite));
                  ScheduleDirectText (COMCODE_INFO_MESSAGE, siBuffer);
                  logPrintf (3, "Multiple Streams: Video %d / Audio %d\n",
                                VideoPids, AudioPids);
               }
               SetupFilterOnEitID (InputDevice, 0, Service->ServiceID);
               InputDevice->State = INPDEV_GET_TV_EIT;
            }
         }
         break;

         case INPDEV_GET_TV_EIT:
         logPrintf (4, "InpDev Pfo GetTVEIT\n");
         {
            struct Channel             *Channel, *chan, *tchan;
            struct Event               *Event;
            struct Descriptor          *Descriptor;

            if (!(Channel = OutputDevice->Channel))
             { logPrintf (2, "no channel defined for EIT\n");
               InputDevice->State = INPDEV_ERROR;
               closeSi (InputDevice); break; }
            Result = siParseEIT (siBuffer);
            xConCat (InputDevice->EventList, Result);
            if (Result && TableHasMoreSections (siBuffer))
               SetupFilterOnEitID (InputDevice,
                   GetSectionNumber(siBuffer)+1, Channel->ServiceID);
            else
            {
               flag = 0;
               xSetRemember (&Channel->LinkRemember);
               xForeach (InputDevice->EventList, Event)
               {
                  if (Event->ServiceID != Channel->ServiceID) continue;
                  if (GetRunningStatus (Event->Status !=
                                     RUNNING_STATUS_RUNNING)) continue;

                  xForeach (Event->Descriptors, Descriptor)
                  {
                     switch (Descriptor->Tag)
                     {
                        case DESCR_LINKAGE:
                           if (((struct LinkageDescriptor *)
                              Descriptor)->LinkageType != 0xB0) break;
                           if (!flag)
                           {
                              if (Channel->LinkRemember)
                                 xMemFreeAll (&Channel->LinkRemember);
                              flag = 1; Channel->LinkList = NULL;
                           }
                           tchan = NULL;
                           if (Channel->LinkList)
                              xForeach (Channel->LinkList, tchan)
                                 if ((tchan->TransportStreamID ==
                                      ((struct LinkageDescriptor *)
                                      Descriptor)->TransportStreamID) &&
                                    (tchan->ServiceID == ((struct
                                      LinkageDescriptor *)
                                      Descriptor)->ServiceID))
                                    break;
                           if (tchan) break;
                           if (chan = getChannelFromLinkageDescriptor
                                     (Descriptor, Channel))
                           {
                              if (!Channel->LinkList)
                                 Channel->LinkList = xNewList (NULL);
                              xAddTail (Channel->LinkList, chan);
                           }
                        break;
                     }
                  }
               }
               if (flag && Channel->LinkList)
               {
                  CloneChannel (chan, Channel, Channel->VideoPid, 0,
                        Channel->AudioPid, 0, screenMsgs[BASIS_CHANNEL]
                                       [OutputDevice->osdLanguage]);
                  xAddTail (Channel->LinkList, chan);
               }
               flag = 0;
               xSetRemember (&Channel->ComponentRemember);
               xForeach (InputDevice->EventList, Event)
               {
                  /* if (Event->ServiceID != Channel->ServiceID) continue; */

                  xForeach (Event->Descriptors, Descriptor)
                  {
                     switch (Descriptor->Tag)
                     {
                        case DESCR_COMPONENT:
                        {
                           struct ComponentInfo *ComponentInfo;

                           if (!flag)
                           {
                              if (Channel->ComponentRemember)
                                 xMemFreeAll(&Channel->ComponentRemember);
                              flag = 1;
                              Channel->ComponentInfos = xNewList (NULL);
                           }
                           CreateComponentInfo (ComponentInfo, xName
                              (Descriptor), ((struct ComponentDescriptor
                              *)Descriptor)->ComponentTag);
                           xAddTail (Channel->ComponentInfos,
                                        ComponentInfo);
                        }
                        break;

                        case DESCR_TIME_SHIFTED_EVENT:
                        {
                           struct tm *StartTime;

                           if (Channel->NvodList)
                              xForeach (Channel->NvodList, chan)
                              {
                                 if ((Event->TransportStreamID ==
                                        chan->TransportStreamID) &&
                                     (Event->ServiceID == chan->ServiceID))
                                 {
                                    StartTime = localtime (&Event->StartTime);
                                    sprintf (siBuffer,
                                       "%s: %2d:%02d",
                                       screenMsgs[TIME_TO_START]
                                       [OutputDevice->osdLanguage],
                                       StartTime->tm_hour,
                                       StartTime->tm_min);
                                    xSetName (chan, siBuffer);
                                    logPrintf (5,
                                        "Starttime %d/%d set to %2d:%02d\n",
                                         Event->TransportStreamID,
                                         Event->ServiceID, StartTime->tm_hour,
                                         StartTime->tm_min);
                                 }
                              }
                        }
                        break;
                     }
                  }
               }
               EventListSize = xSize (InputDevice->EventList);
               InputDevice->EventList = NULL;
               if (Channel->LanguageList && Channel->ComponentInfos)
               {
                  struct ComponentInfo *ComponentInfo;

                  xSetRemember (&Channel->LanguageRemember);
                  xForeach (Channel->LanguageList, chan)
                     xForeach (Channel->ComponentInfos, ComponentInfo)
                        if (chan->AudioTag == ComponentInfo->Tag)
                           xSetName (chan, xName (ComponentInfo));
               }
               if (Channel->NvodList)
               {
                  xForeach (Channel->NvodList, chan)
                  {
                     if (!chan->UsedFlag && xSucc (chan))
                     {
                        chan->UsedFlag = 1;
                        SetupFilterOnEitID (InputDevice,
                            0, chan->ServiceID);
                        break;
                     }
                  }
                  if (chan) break;
               }
               if (Channel->LinkList)
               {
                  logPrintf (6, "Service %d is Linkage\n",
                     InputDevice->Service->ServiceID);
                  sprintf (siBuffer, " \f%c%s\f%c\n",
                     COL2ID(clrGreen), screenMsgs[MORE_CHANNELS_AVAILABLE]
                     [OutputDevice->osdLanguage], COL2ID (clrWhite));
                  ScheduleDirectText (COMCODE_INFO_MESSAGE, siBuffer);
               }
               logPrintf (4, "Programinfo aquisition stopped (%d events)\n",
                             EventListSize);
               InputDevice->State = INPDEV_FINISHED;
               ScheduleDirect (COMCODE_OPEN_EIT);
               closeSi (InputDevice);
            }
         }
         break;
      }
      xSetRemember (OriginalKey);
   }

#undef ImplicitOutputID

   return (0);
}


int getEitEvent (struct InputDevice *InputDevice)
{
   struct OutputDevice                   *OutputDevice;
   struct LIST                           *Result;
   struct MEM_CHUNK                     **OriginalKey;
   unsigned short                         flag, EventListSize;
   int                                    count;
        
   if ((count = read (InputDevice->FileHandle, siBuffer, MAX_SECTION_BUFFER)) < 0)
      return (0);
        
#define ImplicitOutputID OutputDevice->Identifier

   FindOutputDevice (InputDevice, OutputDevice);

   if (OutputDevice)
   {
      OriginalKey = xGetRemember ();
      xSetRemember (&InputDevice->RememberKey);
      switch (InputDevice->State)
      {
         case INPDEV_GET_EIT_ACT:
         logPrintf (4, "InpDev Eit Get_EIT_ACT\n");
         {
            Result = siParseEIT (siBuffer);
            xConCat (InputDevice->EventList, Result);
            if (Result && TableHasMoreSections (siBuffer))
               SetupFilterOnEitID (InputDevice,
                  GetSectionNumber(siBuffer)+1, GetServiceId(siBuffer));
            else
            {
               if (parseEventList (InputDevice) <= 0) InputDevice->UserData--;
               xMemFreeAll (&InputDevice->RememberKey);
               InputDevice->EventList = NULL;
               if (InputDevice->UserData <= 0)
               {
                  InputDevice->UserData =
                     GetGlobalOption("eitAcquisitionNumberOfFails");
                  InputDevice->State = INPDEV_GET_EIT_OTH;
                  SetupFilterAndSection (InputDevice, PID_EIT,
                     TID_EIT_OTH, 0);
               } else SetupFilterAndSection (InputDevice, PID_EIT,
                     TID_EIT_ACT, 0);
            }
         }
         break;

         case INPDEV_GET_EIT_OTH:
         logPrintf (4, "InpDev Eit Get_EIT_OTH\n");
         {
            Result = siParseEIT (siBuffer);
            xConCat (InputDevice->EventList, Result);
            if (Result && TableHasMoreSections (siBuffer))
               SetupFilterOnEitID (InputDevice,
                  GetSectionNumber(siBuffer)+1, GetServiceId(siBuffer));
            else
            {
               if (parseEventList (InputDevice) <= 0) InputDevice->UserData--;
               xMemFreeAll (&InputDevice->RememberKey);
               InputDevice->EventList = NULL;
               if (InputDevice->UserData <= 0)
               {
                  InputDevice->UserData =
                     GetGlobalOption("eitAcquisitionNumberOfFails");
                  InputDevice->State = INPDEV_GET_EIT_ACT_SCH;
                  SetupFilterSectionAndMask (InputDevice, PID_EIT,
                     TID_EIT_ACT_SCH, 0xF0);
               } else SetupFilterAndSection (InputDevice, PID_EIT,
                     TID_EIT_OTH, 0);
            }
         }
         break;

         case INPDEV_GET_EIT_ACT_SCH:
         logPrintf (4, "InpDev Eit Get_EIT_ACT_SCH\n");
         {
            Result = siParseEIT (siBuffer);
            xConCat (InputDevice->EventList, Result);
            if (parseEventList (InputDevice) <= 0) InputDevice->UserData--;
            xMemFreeAll (&InputDevice->RememberKey);
            InputDevice->EventList = NULL;
            if (InputDevice->UserData <= 0)
            {
               InputDevice->UserData =
                  GetGlobalOption("eitAcquisitionNumberOfFails");
               InputDevice->State = INPDEV_GET_EIT_OTH_SCH;
               SetupFilterSectionAndMask (InputDevice, PID_EIT,
                  TID_EIT_OTH_SCH, 0xF0);
            } else SetupFilterSectionAndMask (InputDevice, PID_EIT,
                  TID_EIT_ACT_SCH, 0xF0);
         }
         break;

         case INPDEV_GET_EIT_OTH_SCH:
         logPrintf (4, "InpDev Eit Get_EIT_OTH_SCH\n");
         {
            Result = siParseEIT (siBuffer);
            xConCat (InputDevice->EventList, Result);
            if (parseEventList (InputDevice) <= 0) InputDevice->UserData--;
            xMemFreeAll (&InputDevice->RememberKey);
            InputDevice->EventList = NULL;
            if (InputDevice->UserData <= 0)
            {
               logPrintf (6, "EIT aquisition stopped (%d events)\n",
                          InputDevice->EventList ?
                          xSize (InputDevice->EventList) : 0);
               InputDevice->UserData = 0;
               InputDevice->State = INPDEV_FINISHED;
               closeSi (InputDevice);
            } else SetupFilterSectionAndMask (InputDevice, PID_EIT,
                  TID_EIT_OTH_SCH, 0xF0);
         }
         break;
      }
      xSetRemember (OriginalKey);
   }

#undef ImplicitOutputID

   return (0);
}


int parseEventList (struct InputDevice *InputDevice)
{
   struct MEM_CHUNK          **OriginalKey;
   struct ProgramInfo         *ProgramInfo, *Pi2;
   struct Channel             *Channel, *GlobChannel;
   struct Event               *Event;
   struct Descriptor          *Descriptor;
   struct OutputDevice        *OutputDevice;
   int                         ReturnValue = 0;

   FindOutputDevice (InputDevice, OutputDevice);

   if (!OutputDevice || !(Channel = OutputDevice->Channel))
    { logPrintf (4, "parseEventList called without valid OutputDevice\n");
      return (0); }

   xForeach (InputDevice->EventList, Event)
   {
      xForeach (AvailableChannels, GlobChannel)
         if ((GlobChannel->TransportStreamID == Event->TransportStreamID) &&
             (GlobChannel->ServiceID == Event->ServiceID)) break;

      if (!GlobChannel) continue;

      OriginalKey = xGetRemember ();
      xSetRemember (&GlobChannel->ProgramRemember);

      ProgramInfo = NULL;
      xForeach (GlobChannel->ProgramInfos, ProgramInfo)
         if (ProgramInfo->StartTime == Event->StartTime) break;
      if (ProgramInfo)
      {
         if (ProgramInfo->Content) xMemFree (ProgramInfo->Content);
         if (ProgramInfo->ShortName) xMemFree (ProgramInfo->ShortName);
         if (ProgramInfo->ShortText) xMemFree (ProgramInfo->ShortText);
         if (ProgramInfo->ExtendedName) xMemFree (ProgramInfo->ExtendedName);
         if (ProgramInfo->ExtendedText) xMemFree (ProgramInfo->ExtendedText);
         xRemove (GlobChannel->ProgramInfos, ProgramInfo);
         xMemFree (ProgramInfo);
         ProgramInfo = NULL;
         ReturnValue -= 1;
      }

      xForeach (Event->Descriptors, Descriptor)
      {
         switch (Descriptor->Tag)
         {
            case DESCR_SHORT_EVENT:
            {
               if (!xName(Descriptor) || !xName(Descriptor)[0] ||
                   (Channel->NvodList &&
                    (GetRunningStatus (Event->Status) !=
                      RUNNING_STATUS_RUNNING))) break;

               if (Channel->NvodList && GlobChannel->ProgramInfos &&
                   GlobChannel->ProgramInfos->Head) break;
               logPrintf (4, "ProgramInfo added for %s\n", xName (GlobChannel));
               ReturnValue += 1;
               if (Channel->NvodList)
               {
                  CreateProgramInfo(ProgramInfo,
                     Event->EventID, 0, 2147483647,
                     Event->Status);
               }
               else
               {
                  CreateProgramInfo(ProgramInfo,
                     Event->EventID, Event->StartTime,
                     Event->Duration, Event->Status);
               }
               if (!GlobChannel->ProgramInfos)
                  GlobChannel->ProgramInfos = xNewList (NULL);
               if (!GlobChannel->ProgramInfos->Head)
                  xAddTail (GlobChannel->ProgramInfos,
                               ProgramInfo);
               else
               {
                  xForeachReverse (GlobChannel->ProgramInfos, Pi2)
                     if (Pi2->StartTime <
                        ProgramInfo->StartTime) break;
                  if (!Pi2)
                     xAddHead (GlobChannel->ProgramInfos,
                                  ProgramInfo);
                  else xJoin (ProgramInfo,
                     GlobChannel->ProgramInfos, Pi2);
               }

               ProgramInfo->ShortName =
                  xSetText (xName (Descriptor));
               ProgramInfo->ShortText =
                  xSetText (((struct ShortEventDescriptor
                     *)Descriptor)->Text);
               memcpy (ProgramInfo->LanguageCode, ((struct
                  ShortEventDescriptor *)Descriptor)->
                  LanguageCode, 4);
            }
            break;

            case DESCR_EXTENDED_EVENT:
            {
               struct ExtendedEventItem *Item;

               if (ProgramInfo)
               {
                  if (xName (Descriptor))
                     AddToText (xName (Descriptor),
                        ProgramInfo->ExtendedName);
                  xForeach (((struct ExtendedEventDescriptor*)
                     Descriptor)->Items, Item)
                  {
                     AddItemToText (xName (Item),
                        ProgramInfo->ExtendedText);
                     AddItemToText (Item->Text,
                        ProgramInfo->ExtendedText);
                  }
               }
            }
            break;

            case DESCR_CONTENT:
            {
               int i, j;

               if (ProgramInfo)
               {
                  for (j = 0; j < ((struct ContentDescriptor*)
                     Descriptor)->Amount; j++)
                  {
                     for (i = 0; i < CONTENT_TYPE_NUMBER; i++)
                     if ((GetContentContentNibble1(Descriptor,
                        j) == ContentTypes[i].Nibble1) &&
                         (GetContentContentNibble2(Descriptor,
                        j) == ContentTypes[i].Nibble2))
                        {
                           ProgramInfo->Content = xSetText
                              (ContentTypes[i].Description);
                           break;
                        }
                  }
               }
            }
            break;

            case DESCR_PARENTAL_RATING:
            {
               struct ParentalRating *Rating;

               if (ProgramInfo)
               {
                  xForeach (((struct ParentalRatingDescriptor *)
                              Descriptor)->Ratings, Rating)
                     if (!strncmp (ProgramInfo->LanguageCode,
                             Rating->LanguageCode, 3))
                        ProgramInfo->Rating = Rating->Rating;
               }
            }
            break;
         }
      }
      xSetRemember (OriginalKey);
   }   

   return (ReturnValue);
}
