//////////////////////////////////////////////////////////////
///                                                        ///
/// dvb_vtx.c: routines to handle the VTX stream           ///
///                                                        ///
//////////////////////////////////////////////////////////////

// $Revision: 1.3 $
// $Date: 2001/06/12 18:02:27 $
// $Author: hakenes $
//
//   (C) 2001 Rolf Hakenes <hakenes@hippomi.de>, under the GNU GPL.
//
// dtv is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// dtv is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You may have received a copy of the GNU General Public License
// along with dtv; see the file COPYING.  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#include <ost/dmx.h>

#include "dvb_vtx.h"

#include "definitions.h"
#include "structures.h"
#include "prototypes.h"


static u_char inited = 0;
static u_char lang_char[256];

static u_char *vtxBuffer;
static char sbuf[256];


int openVtx(struct InputDevice *InputDevice)
{
   struct OutputDevice                   *OutputDevice;
   struct dmxPesFilterParams              pesFilterParamsT;
   int                                    ierr = 0;

   if (!inited) { lang_init(); inited = 1; }

   FindOutputDevice (InputDevice, OutputDevice);

   if (!OutputDevice || !OutputDevice->Channel ||
       !OutputDevice->Channel->VtxPid)
    {  logPrintf (0, "no valid output device found for %s\n",
       xName (InputDevice)); return (-1); }
   
   if (InputDevice->FileHandle)
   {
      ioctl (InputDevice->FileHandle, DMX_STOP, 0);
      close (InputDevice->FileHandle);
   }
   if ((InputDevice->FileHandle =
          open(xName (InputDevice), O_RDWR | O_NONBLOCK)) < 0)
    {  logPrintf (0, "cannot open %s (%s)\n", sbuf, strerror (errno));
       return (-1); }

   pesFilterParamsT.pid = OutputDevice->Channel->VtxPid;
   pesFilterParamsT.input = DMX_IN_FRONTEND;
   pesFilterParamsT.output = DMX_OUT_TAP;
   pesFilterParamsT.pesType = DMX_PES_OTHER;
   pesFilterParamsT.flags = 0;

   if (((ierr = ioctl(InputDevice->FileHandle, DMX_SET_PES_FILTER,
                     &pesFilterParamsT)) >= 0) &&
       ((ierr = ioctl(InputDevice->FileHandle, DMX_SET_BUFFER_SIZE,
                     MAX_VTX_BUFFER*8)) >= 0) &&
       ((ierr = ioctl(InputDevice->FileHandle, DMX_START, 0)) >= 0))
   {
      if (OutputDevice->VideoText &&
          (OutputDevice->VideoText->Channel != OutputDevice->Channel))
         OutputDevice->VideoText = NULL;
      logPrintf (3, "Videotext device %s opened\n", xName (InputDevice));
      return (InputDevice->FileHandle);
   }
   logPrintf (0, "set_tpid error (%s)\n", strerror(errno));
   InputDevice->FileHandle = 0;
   close (InputDevice->FileHandle);
   return (-1);
}


void closeVtx(struct InputDevice *InputDevice)
{
   ioctl (InputDevice->FileHandle, DMX_STOP, 0);
   close (InputDevice->FileHandle);
   InputDevice->FileHandle = 0;
   InputDevice->Status = INPDEV_DISABLED;
   InputDevice->State = 0;
   
   return;
}


int getVtxEvent (struct InputDevice *InputDevice)
{
   struct OutputDevice                   *OutputDevice;
   struct MEM_CHUNK                     **OriginalKey;
   int                                    count, i;
        
   if ((count = read (InputDevice->FileHandle, vtxBuffer, MAX_VTX_BUFFER))<0)
      return (0);

   logPrintf (5, "got VtxPage (%d bytes)\n", count);

#define ImplicitOutputID OutputDevice->Identifier

   FindOutputDevice (InputDevice, OutputDevice);

   if (OutputDevice)
   {
      OriginalKey = xGetRemember ();
      xSetRemember (&OutputDevice->VtxRemember);
      if (!OutputDevice->VideoText && OutputDevice->VtxRemember)
         xMemFreeAll (&OutputDevice->VtxRemember);
      if (!OutputDevice->VideoText)
         CreateVideoText (OutputDevice->VideoText, OutputDevice->Channel);
      for (i = 0; i < count; i += 184)
         vtxParse (OutputDevice, &vtxBuffer[i]);
      xSetRemember (OriginalKey);
   }

#undef ImplicitOutputID

   return (0);
}


void vtxParse (struct OutputDevice *OutputDevice, u_char *buf)
{
   int pid, i, j, hdr, mag, mag8, line;
   char *ptr;
   struct VtxPage *VtxPage, *Vtx2;

   for (i = 0;i < 4;i++)
   {
      if (buf[i*46] == 2)
      {
         for (j=(4+i*46);j<(46+i*46);j++)
            { buf[j]=invtab[buf[j]]; }

         hdr = unham16 (&buf[4+i*46]);
         mag = hdr & 0x07; mag8 = mag ?: 8;
         line = (hdr>>3) & 0x1f;
         ptr = &buf[6+i*46];

         switch (line)
         {
            case 0:
            {
               unsigned char b1, b2, b3, b4;
               int pgno, subno, lang, flags, lines, flof;

               b1 = unham16 (ptr);    // page number
               b2 = unham16 (ptr+2);  // subpage number + flags
               b3 = unham16 (ptr+4);  // subpage number + flags
               b4 = unham16 (ptr+6);  // language code + more flags

               pgno = mag8 * 256 + b1;
               subno = (b2 + b3 * 256) & 0x3f7f;
               lang = "\0\4\2\6\1\5\3\7"[b4 >> 5] + (latin1 ? 0 : 8);
               flags = b4 & 0x1f;
               flags |= b3 & 0xc0;
               flags |= (b2 & 0x80) >> 2;
               lines = 1;
               flof = 0;

               conv2latin(ptr + 8, 32, lang);

               if (b1 == 0xff)
                  break;

               enhance (OutputDevice->VideoText->CurrentPage);
               fmt_page (OutputDevice->VideoText->CurrentPage);

               if (((b1 & 0x0F) > 9) || (b1 > 0xF0))
                  break;

               xForeach (OutputDevice->VideoText->Pages, VtxPage)
                  if (VtxPage->PageNumber == pgno) break;
               if (!VtxPage)
               {
                  CreateVtxPage (VtxPage);
                  if (!OutputDevice->VideoText->Pages->Head)
                     xAddTail (OutputDevice->VideoText->Pages, VtxPage);
                  else
                  {
                     xForeachReverse (OutputDevice->VideoText->Pages, Vtx2)
                        if (Vtx2->PageNumber < pgno) break;
                     if (!Vtx2)
                        xAddHead (OutputDevice->VideoText->Pages, VtxPage);
                     else xJoin (VtxPage, OutputDevice->VideoText->Pages, Vtx2);
                  }
               }
               else flags |= PG_CHANGED;
               OutputDevice->VideoText->CurrentPage = VtxPage;
               VtxPage->PageNumber = pgno;
               VtxPage->SubpageNumber = subno;
               VtxPage->LanguageCode = lang;
               VtxPage->Flags = flags | PG_ACTIVE;
               VtxPage->ReceivedLines = lines;
               VtxPage->Flof = 0;

               init_enhance(VtxPage);
               memcpy(VtxPage->Data[0]+0, ptr, 40);
               memset(VtxPage->Data[0]+40, ' ', sizeof(VtxPage->Data)-40);
            }
            break;

            case 1 ... 24:
            {
               if (!(VtxPage = OutputDevice->VideoText->CurrentPage) ||
                   !(OutputDevice->VideoText->CurrentPage->Flags & PG_ACTIVE))
                  break;

               VtxPage->ReceivedLines |= 1 << line;
               conv2latin(ptr, 40, VtxPage->LanguageCode);
               memcpy(VtxPage->Data[line], ptr, 40);
            }
            break;

            case 26:
            {
               int d, t[13];

               if (!(VtxPage = OutputDevice->VideoText->CurrentPage) ||
                   !(OutputDevice->VideoText->CurrentPage->Flags & PG_ACTIVE))
                  break;

               d = unham8(ptr);

               for (i = 0; i < 13; ++i)
                  t[i] = unham24(ptr + 1 + 3*i);

               add_enhance(VtxPage, d, t);
            }
            break;

            case 27:
            {
               // FLOF data (FastText)
               int b1, b2, b3, x;
            
               if (!(VtxPage = OutputDevice->VideoText->CurrentPage) ||
                   !(OutputDevice->VideoText->CurrentPage->Flags & PG_ACTIVE))
                  break;

               b1 = unham8(ptr);
               b2 = unham8(ptr + 37);

               if (b1 != 0 || !(b2 & 8))
                  break;

               for (i = 0; i < 6; ++i)
               {
                  b1 = unham16(ptr+1+6*i);
                  b2 = unham16(ptr+3+6*i);
                  b3 = unham16(ptr+5+6*i);
                  x = (b2 >> 7) | ((b3 >> 5) & 0x06);
                  VtxPage->Link[i].PageNumber = ((mag ^ x) ?: 8) * 256 + b1;
                  VtxPage->Link[i].SubpageNumber = (b2 + b3 * 256) & 0x3f7f;
               }
               VtxPage->Flof = 1;
            }
            break;

            case 30:
            {
               if (mag8 != 8)
                  break;

               ptr[0] = unham8(ptr);      // designation code
               ptr[1] = unham16(ptr+1);   // initial page
               ptr[3] = unham16(ptr+3);   // initial subpage + mag
               ptr[5] = unham16(ptr+5);   // initial subpage + mag

               conv2latin(ptr+20, 20, 0);
            }
            break;

            default:
            break;
         }
      }
   }
}


unsigned char unham8 (unsigned char *p)
{
   unsigned short c1;
  
   c1=unhamtab[p[0]];

   return (c1 & 0x0F);
}

unsigned char unham16 (unsigned char *p)
{
   unsigned short c1,c2;
  
   c1=unhamtab[p[0]];
   c2=unhamtab[p[1]];

   return (c1 & 0x0F) | (c2 & 0x0F) * 16;
}

int unham24(unsigned char *p)
{
    int e, x;

    e = hamm24par[0][p[0]] ^
        hamm24par[1][p[1]] ^
        hamm24par[2][p[2]];
    x = hamm24val[p[0]] +
        p[1] % 128 * 16 +
        p[2] % 128 * 2048;

    // *err += hamm24err[e];
    return x ^ hamm24cor[e];
}


void lang_init (void)
{
   int i;

   memset(lang_char, 0, sizeof(lang_char));
   for (i = 1; i <= 13; i++)
      lang_char[lang_chars[0][i]] = i;
   xMemAlloc (MAX_VTX_BUFFER, &vtxBuffer);
}


void conv2latin (u_char *p, int n, int lang)
{
   int c, gfx = 0;

   while (n--)
   {
      *p = *p & 0x7F;
      if (lang_char[c = *p])
      {
         if (!gfx || (c & 0xa0) != 0x20)
            *p = lang_chars[lang + 1][lang_char[c]];
      }
      else if ((c & 0xe8) == 0)
              gfx = c & 0x10;
      p++;
   }
}

void init_enhance (struct VtxPage *VtxPage)
{
   VtxPage->Enhance.next_des = 0;
}


void add_enhance (struct VtxPage *VtxPage, int dcode, u_int *t)
{
   if (dcode == VtxPage->Enhance.next_des)
   {
      memcpy(VtxPage->Enhance.trip + dcode * 13, t, 13 * sizeof(*t));
      VtxPage->Enhance.next_des++;
   }
   else VtxPage->Enhance.next_des = -1;
}


void enhance (struct VtxPage *VtxPage)
{
   int row = 0;
   unsigned int *p, *e;

   if (!VtxPage || VtxPage->Enhance.next_des < 1)
      return;

   for (p = VtxPage->Enhance.trip,
        e = p + VtxPage->Enhance.next_des * 13; p < e; p++)
      if (*p % 2048 != 2047)
      {
         int adr = *p % 64;
         int mode = *p / 64 % 32;
         int data = *p / 2048 % 128;

         if (adr < 40)
         {
            switch (mode)
            {
               case 15:
                  if (adr < W && row < H)
                     if (latin1)
                        VtxPage->Data[row][adr] = g2map_latin1[data-32];
                     else
                        VtxPage->Data[row][adr] = g2map_latin2[data-32];
               break;

               case 16 ... 31:
                  if (adr < W && row < H)
                  {
                     struct mark *mark = marks + (mode - 16);
                     unsigned char *x;

                     if (x = strchr(mark->g0, data))
                        if (latin1)
                           data = mark->latin1[x - mark->g0];
                        else
                           data = mark->latin2[x - mark->g0];
                     VtxPage->Data[row][adr] = data;
                  }
               break;
            }
         }
         else
         {
            if ((adr -= 40) == 0)
               adr = 24;
                
            switch (mode)
            {
               case 1:
                  row = adr;
               break;
               case 4:
                  row = adr;
               break;
               case 7:
                  if (adr == 23) row = 0;
               break;
            }
         }
      }
}


void fmt_page(struct VtxPage *VtxPage)
{
   char buf[16];
   int x, y;
   unsigned char *p;

   if (!VtxPage) return;

   VtxPage->Double = 0;

   sprintf(buf, "\2%x.%02x\7", VtxPage->PageNumber,
               VtxPage->SubpageNumber & 0xff);

   for (y = 0; y < H; y++)
   {
      struct fmt_char c;
      int last_ch = ' ';
      int dbl = 0, hold = 0;

      p = &VtxPage->Data[y][0];

      c.fg = clrWhite;
      c.bg = clrBlack;
      c.attr = 0;

      for (x = 0; x < W; ++x)
      {
         c.ch = *p++;
         if (y == 0 && x < 8)
            c.ch = buf[x];
         switch (c.ch)
         {
            case 0x00 ... 0x07:        /* alpha + fg color */
               c.fg = (c.ch & 7) + 1;
               c.attr &= ~(EA_GRAPHIC | EA_SEPARATED | EA_CONCEALED);
            goto ctrl;

            case 0x08:                /* flash */
               c.attr |= EA_BLINK;
            goto ctrl;

            case 0x09:                /* steady */
               c.attr &= ~EA_BLINK;
            goto ctrl;

            case 0x0a:                /* end box */
            case 0x0b:                /* start box */
            goto ctrl;

            case 0x0c:                /* normal height */
               c.attr &= EA_DOUBLE;
            goto ctrl;

            case 0x0d:                /* double height */
               c.attr |= EA_DOUBLE;
               dbl = 1;
            goto ctrl;

            case 0x10 ... 0x17:        /* gfx + fg color */
               c.fg = (c.ch & 7) + 1;
               c.attr |= EA_GRAPHIC;
               c.attr &= ~EA_CONCEALED;
            goto ctrl;

            case 0x18:                /* conceal */
               c.attr |= EA_CONCEALED;
            goto ctrl;

            case 0x19:                /* contiguous gfx */
               c.attr &= ~EA_SEPARATED;
            goto ctrl;

            case 0x1a:                /* separate gfx */
               c.attr |= EA_SEPARATED;
            goto ctrl;

            case 0x1c:                /* black bf */
               c.bg = clrBlack;
            goto ctrl;

            case 0x1d:                /* new bg */
               c.bg = c.fg;
            goto ctrl;

            case 0x1e:                /* hold gfx */
               hold = 1;
            goto ctrl;

            case 0x1f:                /* release gfx */
               hold = 0;
            goto ctrl;

            case 0x0e:                /* SO */
            case 0x0f:                /* SI */
            case 0x1b:                /* ESC */
               c.ch = ' ';
            break;

            ctrl:
               c.ch = ' ';
               if (hold && (c.attr & EA_GRAPHIC))
                  c.ch = last_ch;
            break;
         }
         if (c.attr & EA_GRAPHIC)
            if ((c.ch & 0xa0) == 0x20)
            {
               last_ch = c.ch;
               c.ch += (c.ch & 0x40) ? 32 : -32;
            }
         if (c.attr & EA_CONCEALED)
            if (!VtxPage->Reveal)
               c.ch = ' ';
         VtxPage->Page[y][x] = c;
      }
      if (dbl)
      {
         VtxPage->Double |= 1 << y;
         for (x = 0; x < W; ++x)
         {
            if (~VtxPage->Page[y][x].attr & EA_DOUBLE)
               VtxPage->Page[y][x].attr |= EA_HDOUBLE;
            VtxPage->Page[y+1][x] = VtxPage->Page[y][x];
            VtxPage->Page[y+1][x].ch = ' ';
         }
         y++;
      }
   }
   VtxPage->Hidden = VtxPage->Double << 1;
   VtxPage->Flags |= PG_READY;
   VtxPage->Flags &= ~(PG_ACTIVE);

   return;
}


unsigned char *prepare_osd (struct VideoText *VideoText, struct VtxPage *VtxPage)
{
   int x, y, xx, yy, dest_x, dest_y;
   int fg, bg, c, dbl, sep;
   int bitnr, bit, maskbitnr, maskbit;
   unsigned char *bitmap;
   unsigned char* src;

   xMemAlloc (WH*WW, &bitmap); 

   if (!VideoText || !VtxPage)
   {
      memset (bitmap, clrBlack, WH*WW);
      return (bitmap);
   }
   src = (latin1 ? font1_bits : font2_bits);

   if (VideoText->PageNumber)
   {
      VtxPage->Page[0][8].ch = ((VideoText->PageNumber & 0xF00) >> 8) + '0';
      VtxPage->Page[0][9].ch = ((VideoText->PageNumber & 0x0F0) >> 4) + '0';
      VtxPage->Page[0][10].ch = ((VideoText->PageNumber & 0x00F) >> 0) + '0';
      if (VtxPage->Page[0][8].ch == '0')
      { VtxPage->Page[0][8].ch = '.';
        if (VtxPage->Page[0][9].ch == '0') VtxPage->Page[0][9].ch = '.'; }
   }

   for (y = 0; y < H; ++y)
   {
      for (x = 0; x < W; ++x)
      { 
         fg = VtxPage->Page[y][x].fg;
         bg = VtxPage->Page[y][x].bg;
         c = VtxPage->Page[y][x].ch;
         dbl = (VtxPage->Page[y][x].attr & EA_DOUBLE) ? 1 : 0;
         sep = (VtxPage->Page[y][x].attr & EA_SEPARATED) ? 1 : 0;
         dest_x = x * CW;
         dest_y = y * CH;

         if (VtxPage->Double & (1<<(y-1)))
         {
            if (VtxPage->Page[y-1][x].attr & EA_HDOUBLE)
            {
               for (yy = 0; yy < CH; yy++)
               {
                  for (xx = 0; xx < CW; xx++)
                  {
                     bitnr = (c/32*CH + yy)*CW*32+ c%32*CW +xx;
                     bit = (*(src+bitnr/8))&(1<<bitnr%8);
                     if (sep)
                     {
                        maskbitnr=(0xa0/32*CH + yy)*CW*32+
                                   0xa0%32*CW +xx;
                        maskbit=(*(src+maskbitnr/8))&(1<<maskbitnr%8);
                        *(bitmap+WW*(dest_y+yy)+dest_x+xx)=
                           (char)((bit && (!maskbit)) ? fg : bg);
                     }
                     else 
                        *(bitmap+WW*(dest_y+yy)+dest_x+xx)=
                           (char)(bit ? fg : bg);
                  }
               }
            }
         }
         else
         {
            for (yy = 0; yy < (CH<<dbl); yy++)
            {
               for (xx = 0; xx < CW; xx++)
               {
                  bitnr = (c/32*CH + (yy>>dbl))*CW*32+ c%32*CW +xx;
                  bit = (*(src+bitnr/8))&(1<<bitnr%8);
                  if (sep)
                  {
                     maskbitnr=(0xa0/32*CH + (yy>>dbl))*CW*32+
                                0xa0%32*CW +xx;
                     maskbit=(*(src+maskbitnr/8))&(1<<maskbitnr%8);
                     *(bitmap+WW*(dest_y+yy)+dest_x+xx)=
                        (char)((bit && (!maskbit)) ? fg : bg);
                  }
                  else 
                     *(bitmap+WW*(dest_y+yy)+dest_x+xx)=
                        (char)(bit ? fg : bg);
               }
            }
         }
      }
   }
   return (bitmap);
}


unsigned char *prepare_osd_header (struct VideoText *VideoText, struct VtxPage *VtxPage)
{
   int x, y, xx, yy, dest_x, dest_y;
   int fg, bg, c, dbl, sep;
   int bitnr, bit, maskbitnr, maskbit, curSub, curPage;
   unsigned char *bitmap;
   unsigned char* src;
   struct fmt_char  Page[1][40];
   time_t                  ThisTime;
   struct tm              *StartTime;

   xMemAlloc (CH*WW, &bitmap); 

   if (!VideoText)
   {
      memset (bitmap, clrBlack, CH*WW);
      return (bitmap);
   }
   src = (latin1 ? font1_bits : font2_bits);

   if (VtxPage && (VtxPage == VideoText->DisplayedPage))
   {
      curPage = VtxPage->PageNumber;
      curSub = VtxPage->SubpageNumber;
   }
   else
   {
      xForeach (VideoText->Pages, VtxPage)
         if (VtxPage->Flags & PG_ACTIVE)
         {
            curPage = VtxPage->PageNumber;
            curSub = VtxPage->SubpageNumber;
            break;
         }
      if (!VtxPage)
      {
         memset (bitmap, clrBlack, CH*WW);
         return (bitmap);
      }
   }
   xForeach (VideoText->Pages, VtxPage)
      if (VtxPage->Flags & PG_READY)
      {
         memcpy (Page, VtxPage->Page, sizeof(Page));
         break;
      }
   if (!VtxPage)
   {
      memset (bitmap, clrBlack, CH*WW);
      return (bitmap);
   }
   Page[0][1].ch = ((curPage & 0xF00) >> 8) + '0';
   Page[0][2].ch = ((curPage & 0x0F0) >> 4) + '0';
   Page[0][3].ch = ((curPage & 0x00F) >> 0) + '0';
   Page[0][5].ch = ((curSub & 0xF0) >> 4) + '0';
   Page[0][6].ch = ((curSub & 0x0F) >> 0) + '0';

   if (VideoText->PageNumber)
   {
      Page[0][8].ch = ((VideoText->PageNumber & 0xF00) >> 8) + '0';
      Page[0][9].ch = ((VideoText->PageNumber & 0x0F0) >> 4) + '0';
      Page[0][10].ch = ((VideoText->PageNumber & 0x00F) >> 0) + '0';
      if (Page[0][8].ch == '0')
      { Page[0][8].ch = '.'; if (Page[0][9].ch == '0') Page[0][9].ch = '.'; }
   }
   else
   {
      Page[0][8].ch = ((curPage & 0xF00) >> 8) + '0';
      Page[0][9].ch = ((curPage & 0x0F0) >> 4) + '0';
      Page[0][10].ch = ((curPage & 0x00F) >> 0) + '0';
   }

   ThisTime = time (NULL);
   StartTime = localtime (&ThisTime);

   Page[0][32].ch = (StartTime->tm_hour/10) + '0'; // hour-ten
   Page[0][33].ch = (StartTime->tm_hour%10) + '0'; // hour-one
   Page[0][35].ch = (StartTime->tm_min/10) + '0'; // minute-ten
   Page[0][36].ch = (StartTime->tm_min%10) + '0'; // minute-one
   Page[0][38].ch = (StartTime->tm_sec/10) + '0'; // second-ten
   Page[0][39].ch = (StartTime->tm_sec%10) + '0'; // second-one

   for (y = 0; y < 1; ++y)
   {
      for (x = 0; x < W; ++x)
      { 
         fg = Page[y][x].fg;
         bg = Page[y][x].bg;
         c = Page[y][x].ch;
         dbl = (Page[y][x].attr & EA_DOUBLE) ? 1 : 0;
         sep = (Page[y][x].attr & EA_SEPARATED) ? 1 : 0;
         dest_x = x * CW;
         dest_y = y * CH;

         for (yy = 0; yy < (CH<<dbl); yy++)
         {
            for (xx = 0; xx < CW; xx++)
            {
               bitnr = (c/32*CH + (yy>>dbl))*CW*32+ c%32*CW +xx;
               bit = (*(src+bitnr/8))&(1<<bitnr%8);
               if (sep)
               {
                  maskbitnr=(0xa0/32*CH + (yy>>dbl))*CW*32+
                             0xa0%32*CW +xx;
                  maskbit=(*(src+maskbitnr/8))&(1<<maskbitnr%8);
                  *(bitmap+WW*(dest_y+yy)+dest_x+xx)=
                     (char)((bit && (!maskbit)) ? fg : bg);
               }
               else 
                  *(bitmap+WW*(dest_y+yy)+dest_x+xx)=
                     (char)(bit ? fg : bg);
            }
         }
      }
   }
   return (bitmap);
}
