//////////////////////////////////////////////////////////////
///                                                        ///
/// input.c: input functions of dtv                        ///
///                                                        ///
//////////////////////////////////////////////////////////////

// $Revision: 1.8 $
// $Date: 2001/06/13 16:28:45 $
// $Author: hakenes $
//
//   (C) 2001 Rolf Hakenes <hakenes@hippomi.de>, under the GNU GPL.
//
// dtv is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// dtv is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You may have received a copy of the GNU General Public License
// along with dtv; see the file COPYING.  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <termios.h>

#include <dcl.h>

#include "definitions.h"
#include "structures.h"
#include "prototypes.h"


static char requestBuffer[MAX_MESSAGE_LEN];

static struct termios keyboardTty;
static int keyboardIsRaw = 0;

int openKeyboard(struct InputDevice *InputDevice)
{
   static struct termios term;

   tcgetattr(0, &keyboardTty);

   term = keyboardTty;
   term.c_iflag = IGNPAR | IGNBRK;
   term.c_oflag = 0;
   term.c_cflag = CREAD | CS8;
   term.c_lflag &= ~(ICANON | ECHO | ISIG);
   term.c_cc[VTIME] = 0;
   term.c_cc[VMIN] = 0;
   tcsetattr(0, TCSANOW, &term);

   keyboardIsRaw = 1;

   return (0);
}

void closeKeyboard(struct InputDevice *InputDevice)
{
   if (keyboardIsRaw != 0) {
      tcsetattr(0, TCSANOW, &keyboardTty);
      keyboardIsRaw = 0;
   }
   InputDevice->Status = INPDEV_DISABLED;
}

int openRemote(struct InputDevice *InputDevice)
{
   if (!xName(InputDevice) ||
      (InputDevice->FileHandle = open(xName(InputDevice), O_RDWR)) < 0)
      return (-1);

   return (InputDevice->FileHandle);
}

void closeRemote(struct InputDevice *InputDevice)
{
   if (InputDevice->FileHandle)
   {
      close (InputDevice->FileHandle);
      InputDevice->FileHandle = 0;
   }
   InputDevice->Status = INPDEV_DISABLED;
   return;
}

int openLircd(struct InputDevice *InputDevice)
{
   struct sockaddr_un addr;

   if (!xName(InputDevice))
      return (-1);

   addr.sun_family = AF_UNIX;
   strcpy(addr.sun_path, xName(InputDevice));

   InputDevice->Status = INPDEV_DISABLED;

   if ((InputDevice->FileHandle = socket(AF_UNIX, SOCK_STREAM, 0)) >= 0)
   {
      if (connect(InputDevice->FileHandle, (struct sockaddr *)&addr,
         sizeof(addr)) >= 0)
      {
         InputDevice->Status = INPDEV_RUNNING;
         return (InputDevice->FileHandle);
      }
      close(InputDevice->FileHandle);
   }
   InputDevice->FileHandle = 0;
   return (-1);
}

void closeLircd(struct InputDevice *InputDevice)
{
   if (InputDevice->FileHandle)
   {
      close (InputDevice->FileHandle);
      InputDevice->FileHandle = 0;
   }
   InputDevice->Status = INPDEV_DISABLED;
   return;
}

int openTcpip(struct InputDevice *InputDevice)
{
   struct sockaddr_in    inet_addr;

   if (!xName(InputDevice)) return (-1);

   if ((InputDevice->FileHandle = socket (PF_INET, SOCK_STREAM, 0)) < 0)
    { logPrintf (0, "server socket error\n"); return (-1); }

   memset (&inet_addr, 0, sizeof (inet_addr));
   inet_addr.sin_family = AF_INET;
   inet_addr.sin_port = htons (atoi(xName(InputDevice)));

   if (bind (InputDevice->FileHandle, (struct sockaddr_in *) &inet_addr,
          sizeof (struct sockaddr_in)) < 0)
    { logPrintf (0, "server bind error\n"); return (-1); }

   if (listen (InputDevice->FileHandle, 5) < 0)
    { logPrintf (0, "server listen error\n"); return (-1); }

   logPrintf (4, "tcpip server successfully opened at port %s\n",
                     xName(InputDevice));
   InputDevice->Status = INPDEV_RUNNING;
   return (InputDevice->FileHandle);
}

void closeTcpip(struct InputDevice *InputDevice)
{
   if (InputDevice->FileHandle)
   {
      close (InputDevice->FileHandle);
      InputDevice->FileHandle = 0;
   }
   InputDevice->Status = INPDEV_DISABLED;
   return;
}


int getRemoteEvent (struct InputDevice *InputDevice)
{
   int                    data;
   struct OutputDevice   *OutputDevice;
   struct Implication    *Implication;
   struct TranslateItem  *TranslateItem;
   int                    c;

   read (InputDevice->FileHandle, &data, sizeof (data));

   logPrintf (6, "RemoteEvent: %04X\n", data);

   xForeach (InputDevice->Implications, Implication)
      xForeach (OutputDevices, OutputDevice)
         if (OutputDevice->Identifier == Implication->OutputID)
         {
            if (Implication->TranslationTable)
            {
               c = 0;
               xForeach (Implication->TranslationTable, TranslateItem)
                  if (TranslateItem->Value == data)
                   { c = TranslateItem->CtrlChar; break; }
            }
            else c = data;
            processInputEvent (OutputDevice, c);
         }

   return (0);
}


int getLircdEvent (struct InputDevice *InputDevice)
{
   struct OutputDevice   *OutputDevice;
   struct Implication    *Implication;
   struct TranslateItem  *TranslateItem;
   int                    c, Length, i, k;
   char                  *requestPtr, *repeatPtr;

   memset (requestBuffer, 0, MAX_MESSAGE_LEN);
   if ((Length = read (InputDevice->FileHandle, &requestBuffer, MAX_MESSAGE_LEN)) < 0)
      return (0);

   while ((requestBuffer[strlen(requestBuffer)-1] == '\n') ||
          (requestBuffer[strlen(requestBuffer)-1] == '\r'))
      requestBuffer[strlen(requestBuffer)-1] = '\0';

   logPrintf (6, "LircdEvent: %s\n", requestBuffer);

   for (i = strlen(requestBuffer)-1, k = 2; i > 0; i--)
      if ((requestBuffer[i] == ' ') && !(--k))
         break;
   requestPtr = &requestBuffer[i+1];
   requestBuffer[i] = '\0';
   for (i = strlen(requestBuffer)-1; i > 0; i--)
      if (requestBuffer[i] == ' ')
         break;
   repeatPtr = &requestBuffer[i+1];
   requestBuffer[i] = '\0';

   logPrintf (6, "RequestPtr: %s/%d\n", requestPtr, atoi(repeatPtr));

   if (atoi(repeatPtr) != 0) return (0);

   xForeach (InputDevice->Implications, Implication)
      xForeach (OutputDevices, OutputDevice)
         if (OutputDevice->Identifier == Implication->OutputID)
         {
            c = 0;
            xForeach (Implication->TranslationTable, TranslateItem)
               if (!strcasecmp (TranslateItem->Command, requestPtr))
                { c = TranslateItem->CtrlChar; break; }
            processInputEvent (OutputDevice, c);
         }

   return (0);
}


int getKeyboardEvent (struct InputDevice *InputDevice)
{
   char                   data;
   struct OutputDevice   *OutputDevice;
   struct Implication    *Implication;
   struct TranslateItem  *TranslateItem;
   int                    c, count;

   if ((count = read (InputDevice->FileHandle, &data, sizeof (data))) < 0)
      return (0);

   xForeach (InputDevice->Implications, Implication)
      xForeach (OutputDevices, OutputDevice)
         if (OutputDevice->Identifier == Implication->OutputID)
         {
            if (Implication->TranslationTable)
            {
               c = 0;
               xForeach (Implication->TranslationTable, TranslateItem)
                  if (TranslateItem->Value == data)
                   { c = TranslateItem->CtrlChar; break; }
            }
            else c = data;
            processInputEvent (OutputDevice, c);
         }

   return (0);
}


int getTcpipEvent (struct InputDevice *InputDevice)
{
   struct sockaddr_in    inet_addr;
   int                   ClientFD, Length, id = 0;
   struct InputDevice   *NewDevice;

   memset (&inet_addr, 0, sizeof (inet_addr));
   inet_addr.sin_family = AF_INET;
   inet_addr.sin_port = htons (atoi(xName(InputDevice)));
   Length = sizeof (inet_addr);

   if ((ClientFD = accept (InputDevice->FileHandle,
          (struct sockaddr *) &inet_addr, &Length)) < 0)
    { logPrintf (0, "server accept error (%s)\n",
      strerror (errno)); return (-1); }

   xForeach (InputDevices, NewDevice)
      if (!strcmp (xName (NewDevice), "tcpipClient") &&
          NewDevice->Status == INPDEV_DISABLED) break;
   if (!NewDevice)
   {
      xForeach (InputDevices, NewDevice)
         if (NewDevice->Identifier >= id)
            id = NewDevice->Identifier + 1;
      CreateInputDevice (NewDevice, "tcpipClient", id, NULL,
         closeTcpip, getTcpipClientEvent);
      NewDevice->Type = INPTYP_POLLING;
      xAddTail (InputDevices, NewDevice);
   }
   NewDevice->Status = INPDEV_RUNNING;
   NewDevice->FileHandle = ClientFD;

   logPrintf (4, "new client accepted at port %s\n", xName(InputDevice));
   return (0);
}


int getTcpipClientEvent (struct InputDevice *InputDevice)
{
   int                   Length;

   memset (requestBuffer, 0, MAX_MESSAGE_LEN);
   if ((Length = read (InputDevice->FileHandle, &requestBuffer, MAX_MESSAGE_LEN)) < 0)
      return (0);

   if (!Length)
   {
      closeTcpip (InputDevice);
      logPrintf (4, "client lost from port %s\n", xName(InputDevice));
      return (0);
   }
   while ((requestBuffer[strlen(requestBuffer)-1] == '\n') ||
          (requestBuffer[strlen(requestBuffer)-1] == '\r'))
      requestBuffer[strlen(requestBuffer)-1] = '\0';
   if (!strcasecmp (requestBuffer, "exit"))
   {
      closeTcpip (InputDevice);
      logPrintf (4,"client closed connection at port %s\n",xName(InputDevice));
      return (0);
   }
   logPrintf (4, "received: %s\n", requestBuffer);
   Dcl_Eval (interp, requestBuffer);
   if (interp->result && interp->result[0])
   {
      logPrintf (4, "answered: %s\n", interp->result);
      write (InputDevice->FileHandle, interp->result, strlen(interp->result));
      write (InputDevice->FileHandle, "\n", 1);
   }
   return (0);
}


int processInputEvent (struct OutputDevice *OutputDevice, int c)
{
   unsigned short         CommandCode;
   struct Channel        *Channel;
   struct Command        *Command;
   int                    CmdVal;

   logPrintf (6, "InputEvent: %04X\n", c);

#define ImplicitOutputID OutputDevice->Identifier

   switch (OutputDevice->State)
   {
      case OUTST_OFF:
         ScheduleDirect (COMCODE_ON);
      break;

      case OUTST_IDLE:
         switch (c)
         {
            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
               FindCommand (Command, COMCODE_CHANNEL_SET);
               if (Command)
               {
                  if (Command->Value < 100)
                     CmdVal = Command->Value*10+(c-'0');
                  else CmdVal = c - '0';
                  Command->Value = CmdVal;
                  gettimeofday (&Command->TimeStamp, NULL);
               }
               else
               {
                  CmdVal = c - '0';
                  ScheduleDelayedValue (COMCODE_CHANNEL_SET, CmdVal, 10);
               }
               ScheduleDirectValue (COMCODE_CHANNEL_INFO, CmdVal);
            break;

            case '-':
               ScheduleDirect (COMCODE_CHANNEL_DOWN);
            break;

            case '+':
               ScheduleDirect (COMCODE_CHANNEL_UP);
            break;

            case 'p':
            break;

            case 'o':
               ScheduleDirectValue (COMCODE_CHANNEL_INFO, -1);
               ScheduleDirectValue (COMCODE_PROGRAM_INFO, 5);
            break;

            case 'y':
               ScheduleDirect (COMCODE_ENTER_LANGUAGE_MENU);
            break;

            case 'g':
               ScheduleDirect (COMCODE_ENTER_NVOD_MENU);
               ScheduleDirect (COMCODE_ENTER_LINK_MENU);
            break;

            case 'b':
               ScheduleDirect (COMCODE_OPEN_VIDEOTEXT);
            break;

            case 'r':
               ScheduleDirect (COMCODE_ENTER_USER_MENU);
            break;

            case 'q':
               ScheduleDirect (COMCODE_QUIT_PROGRAM);
            break;

            case 'a':
               ScheduleDirect (COMCODE_ENTER_CHANNEL_MENU);
            break;
         }
      break;

      case OUTST_CHANNEL_MENU:
         switch (c)
         {
            case '-':
               ScheduleDirect (COMCODE_CHANNEL_MENU_DOWN);
            break;

            case '+':
               ScheduleDirect (COMCODE_CHANNEL_MENU_UP);
            break;

            case 'r': case 'g': case 'b': case 'y': case 'o':
               ScheduleDirect (COMCODE_USE_CHANNEL_MENU);
            break;

            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
            case 'a':
               ScheduleDirect (COMCODE_QUIT_CHANNEL_MENU);
            break;
         }
      break;

      case OUTST_LANGUAGE_MENU:
         switch (c)
         {
            case '-':
               ScheduleDirect (COMCODE_LANGUAGE_MENU_DOWN);
            break;

            case '+':
               ScheduleDirect (COMCODE_LANGUAGE_MENU_UP);
            break;

            case 'o':
               ScheduleDirect (COMCODE_USE_LANGUAGE_MENU);
            break;

            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
            case 'y':
               ScheduleDirect (COMCODE_QUIT_LANGUAGE_MENU);
            break;
         }
      break;

      case OUTST_NVOD_MENU:
         switch (c)
         {
            case '-':
               ScheduleDirect (COMCODE_NVOD_MENU_DOWN);
            break;

            case '+':
               ScheduleDirect (COMCODE_NVOD_MENU_UP);
            break;

            case 'o':
               ScheduleDirect (COMCODE_USE_NVOD_MENU);
            break;

            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
            case 'g':
               ScheduleDirect (COMCODE_QUIT_NVOD_MENU);
            break;
         }
      break;

      case OUTST_USER_MENU:
         switch (c)
         {
            case '-':
               ScheduleDirect (COMCODE_USER_MENU_DOWN);
            break;

            case '+':
               ScheduleDirect (COMCODE_USER_MENU_UP);
            break;

            case 'o':
               ScheduleDirect (COMCODE_USE_USER_MENU);
            break;

            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
            case 'r':
               ScheduleDirect (COMCODE_QUIT_USER_MENU);
            break;
         }
      break;

      case OUTST_VIDEOTEXT:
         switch (c)
         {
            case '0': case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
               FindCommand (Command, COMCODE_SHOW_VIDEOTEXT);
               if (Command)
               {
                  if (Command->Value < 0x10000)
                     CmdVal = Command->Value*0x10+((c-'0')*0x100);
                  else CmdVal = (c - '0') * 0x100;
                  Command->Value = CmdVal;
                  gettimeofday (&Command->TimeStamp, NULL);
               }
               else
               {
                  CmdVal = (c - '0') * 0x100;
                  ScheduleDirectValue (COMCODE_SHOW_VIDEOTEXT, CmdVal);
               }
            break;

            case '-':
               ScheduleDirect (COMCODE_VTX_PAGE_DOWN);
            break;

            case '+':
               ScheduleDirect (COMCODE_VTX_PAGE_UP);
            break;

            case 'r': case 'g': case 'b': case 'y': case 'o':
               ScheduleDirect (COMCODE_QUIT_VIDEOTEXT);
            break;
         }
      break;
   }

#undef ImplicitOutputID

   return (0);
}
