//////////////////////////////////////////////////////////////
///                                                        ///
/// scheduler.c: main scheduling functions of dtv          ///
///                                                        ///
//////////////////////////////////////////////////////////////

// $Revision: 1.4 $
// $Date: 2001/06/13 14:12:31 $
// $Author: hakenes $
//
//   (C) 2001 Rolf Hakenes <hakenes@hippomi.de>, under the GNU GPL.
//
// dtv is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// dtv is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You may have received a copy of the GNU General Public License
// along with dtv; see the file COPYING.  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.

#include <poll.h>

#include "definitions.h"
#include "structures.h"
#include "prototypes.h"



int execScheduler (void)
{
   collectInputEvents ();
   return (processCommands ());
}


int collectInputEvents (void)
{
   struct InputDevice     *InputDevice;
   struct Implication     *Implication;
   struct pollfd           pfd;
   int                     Result = 0;

   xForeach (InputDevices, InputDevice)
      if (InputDevice->Status == INPDEV_RUNNING)
      {
         if (InputDevice->Type & INPTYP_POLLING)
         {
            pfd.fd = InputDevice->FileHandle;
            pfd.events = POLLIN;
            if (poll (&pfd, 1, 0) > 0)
               Result |= InputDevice->GetEvent (InputDevice);
         }
         else Result |= InputDevice->GetEvent (InputDevice);
      }
   return (Result);
}


int openInputDevices (void)
{
   struct InputDevice     *InputDevice;
   struct Implication     *Implication;
   int                     Result = 0;

   xForeach (InputDevices, InputDevice)
   {
      if (!xSize (InputDevice->Implications))
      {
         if ((InputDevice->Type & INPTYP_ACTIVE) &&
               (InputDevice->Status != INPDEV_RUNNING))
         {
            if (InputDevice->Open &&
                InputDevice->Open(InputDevice) >= 0)
               InputDevice->Status = INPDEV_RUNNING;
            else Result |= InputDevice->FileHandle;
         }
      }
      else
      {
         xForeach (InputDevice->Implications, Implication)
         {
            if (!(Implication->Type & IMPL_DEPENDENT) &&
                  (InputDevice->Status != INPDEV_RUNNING))
            {
               if (InputDevice->Open &&
                   InputDevice->Open(InputDevice) >= 0)
                  InputDevice->Status = INPDEV_RUNNING;
               else Result |= InputDevice->FileHandle;
            }
         }
      }
   }
   return (Result);
}

int closeInputDevices (void)
{
   struct InputDevice     *InputDevice;

   xForeach (InputDevices, InputDevice)
      if (InputDevice->Status == INPDEV_RUNNING)
      {
         InputDevice->Close (InputDevice);
         InputDevice->Status = INPDEV_DISABLED;
      }
   return (0);
}

int closeOutputDevices (void)
{
   struct OutputDevice     *OutputDevice;

   xForeach (OutputDevices, OutputDevice)
      OutputDevice->Close (OutputDevice);

   return (0);
}

int processCommands (void)
{
   struct Command         *Command, *NextCommand;
   struct OutputDevice    *OutputDevice;
   int                     Result = 0;
   struct timeval          CurrentTime;

   if (xAllocatedMemory > MAX_MEMORY_USAGE)
    { logPrintf (0, "Maximum allowed memory exceeded\n"); }

   xForeach (OutputDevices, OutputDevice)
      if (Command = (struct Command *) CommandQueue->Head)
      do
      {
         NextCommand = (struct Command *) xSucc (Command);
         if (Command->OutputID == OutputDevice->Identifier)
         {
            logPrintf (8, "Command: %d %d %d %d %d %d %d\n", Command->Code,
                       Command->Status, Command->Type, Command->Value,
                       Command->OutputID, Command->TimeStamp.tv_sec,
                       Command->TimeOut);
            switch (Command->Type)
            {
               case COMTYPE_DIRECT:
                  Result |= OutputDevice->Process (OutputDevice, Command);
                  NextCommand = (struct Command *) xSucc (Command);
                  FreeCommand (Command);
               break;

               case COMTYPE_DELAYED:
                  gettimeofday (&CurrentTime, NULL);
                  if (CommandHasTimedOut (Command, CurrentTime))
                  {
                     Result |= OutputDevice->Process (OutputDevice, Command);
                     NextCommand = (struct Command *) xSucc (Command);
                     FreeCommand (Command);
                  }
               break;
            }
         }
      } while (Command = NextCommand);

   return (Result);
}
