//////////////////////////////////////////////////////////////
///                                                        ///
/// dvb_front.c: routines dealing with DVB in dtv_scan     ///
///                                                        ///
//////////////////////////////////////////////////////////////

// $Revision: 1.9 $
// $Date: 2001/06/23 16:33:40 $
// $Author: hakenes $
//
//   (C) 2001 Rolf Hakenes <hakenes@hippomi.de>, under the GNU GPL.
//
// dtv_scan is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// dtv_scan is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You may have received a copy of the GNU General Public License
// along with dtv_scan; see the file COPYING.  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.

#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>

#include <ost/dmx.h>
#include <ost/frontend.h>
#include <ost/sec.h>
#include <ost/video.h>

#include <libsi.h>

#include "definitions.h"
#include "structures.h"
#include "prototypes.h"


static char sbuf[4096];

static int NumChannels;

static struct LIST *TmpChanList = NULL;
static struct MEM_CHUNK *TmpChanListRemember = NULL;

int openDvbDevice (struct OutputDevice *OutputDevice)
{
   if (!xName (OutputDevice)) return (-1);

   return (0);
}


void closeDvbDevice (struct OutputDevice *OutputDevice)
{
   if (!OutputDevice->audioDev) return;
   ioctl(OutputDevice->audioDev, DMX_STOP, 0);
   close (OutputDevice->audioDev);
   if (!OutputDevice->videoDev) return;
   ioctl(OutputDevice->videoDev, DMX_STOP, 0);
   close (OutputDevice->videoDev);

   return;
}


int setDvbChannel (struct OutputDevice *OutputDevice, struct Transponder *Transponder)
{
   int                          fd_frontend, fd_sec, newFrequency;
   struct dmxPesFilterParams    pesFilterParams;
   int                          i = 2, freq, count, res;
   feStatus                     stat = 0;

   if (!OutputDevice || !Transponder)
    {  logPrintf (0, "setDvbChannel called with invalid parameters\n");
       return (-1); }

   if (Transponder->Polarisation == 'c' || Transponder->Polarisation == 'C')
   {
      struct qamParameters qam;
      struct qamEvent event;

      sprintf(sbuf, "/dev/ost/qamfe%s", xName(OutputDevice));
      if ((fd_frontend=open(sbuf, O_RDWR)) < 0)
       {  logPrintf (0, "cannot open %s (%s)\n", sbuf, strerror (errno));
          return (-1); }

      qam.FEC_inner = 0 /* FEC_AUTO */;
      qam.SymbolRate = Transponder->SymbolRate * 1000;
      qam.QAM = 2 /* QAM_64 */;
      qam.Frequency = Transponder->Frequency * 1000; /* this is due to the fact,
                                                    that frequencies for QAM
                                                    have to be specified in Hz,
                                                    for QPSK in kHz...
                                                    why the hell is that ?? */
      if (ioctl(fd_frontend, QAM_TUNE, &qam) <0)
       { logPrintf (0, "setfront qamfe error (%s)\n", strerror (errno));
         close (fd_frontend); return (-1); }

      if ((res = ioctl(fd_frontend, QAM_GET_EVENT, &event)) == -EBUFFEROVERFLOW)
         res = ioctl(fd_frontend, QAM_GET_EVENT, &event);
      if (res < 0)
       { logPrintf (0, "get event qamfe error (%s)\n", strerror (errno));
         close (fd_frontend); return (-1); }

      if (event.type != FE_COMPLETION_EV)
       { logPrintf (0, "could not tune to %d/%c/%d\n", Transponder->Frequency,
            Transponder->Polarisation, Transponder->SymbolRate);
            close (fd_frontend); return (-1); }

      newFrequency = event.u.completionEvent.Frequency/1000; /* see above */
      if (newFrequency != Transponder->Frequency)
         logPrintf (4, "frequency adjusted from %ld to %ld\n",
             Transponder->Frequency, newFrequency);

      close (fd_frontend);
   }
   else
   {
      struct qpskParameters qpsk;
      struct qpskEvent event;
      struct secCommand scmd;
      struct secCmdSequence scmds;
      int DiseqcNr, LnbLsof, LnbFreqLow, LnbFreqHigh;

      if (Transponder->Satellite)
      {
         DiseqcNr = Transponder->Satellite->DiseqcNr;
         LnbLsof = Transponder->Satellite->LnbLsof;
         LnbFreqLow = Transponder->Satellite->LnbFreqLow;
         LnbFreqHigh = Transponder->Satellite->LnbFreqHigh;
      }
      else
      {
         DiseqcNr = GetGlobalOption ("dvbDefaultDiseqcNr");
         LnbLsof = GetGlobalOption ("dvbDefaultLnbLsof");
         LnbFreqLow = GetGlobalOption ("dvbDefaultLnbFreqLow");
         LnbFreqHigh = GetGlobalOption ("dvbDefaultLnbFreqHigh");
      }
      LnbLsof *= 1000; LnbFreqLow*= 1000; LnbFreqHigh *= 1000;

      sprintf(sbuf, "/dev/ost/qpskfe%s", xName(OutputDevice));
      if ((fd_frontend=open(sbuf, O_RDWR)) < 0)
       {  logPrintf (0, "cannot open %s (%s)\n", sbuf, strerror (errno));
          return (-1); }

      sprintf(sbuf, "/dev/ost/sec%s", xName(OutputDevice));
      if ((fd_sec=open(sbuf, O_RDWR)) < 0)
       {  logPrintf (0, "cannot open %s (%s)\n", sbuf, strerror (errno));
          close (fd_frontend); return (-1); }

      qpsk.FEC_inner = 0 /* FEC_AUTO */;
      qpsk.SymbolRate = Transponder->SymbolRate * 1000;

      if (Transponder->Frequency < LnbLsof)
      {
         qpsk.iFrequency = Transponder->Frequency - LnbFreqLow;
         scmds.continuousTone = SEC_TONE_OFF;
      }
      else
      {
         qpsk.iFrequency = Transponder->Frequency - LnbFreqHigh;
         scmds.continuousTone = SEC_TONE_ON;
      }
      if (Transponder->Polarisation == 'h' || Transponder->Polarisation == 'H')
         scmds.voltage = SEC_VOLTAGE_18;
      else scmds.voltage = SEC_VOLTAGE_13;

      /* set_diseqc_nb(scmd, scmds, nr); */
      scmd.type=0;
      scmd.u.diseqc.addr = 0x10;
      scmd.u.diseqc.cmd = 0x38;
      scmd.u.diseqc.numParams = 1;
      scmd.u.diseqc.params[0] = 0xF0 | ((DiseqcNr * 4) & 0x0F) |
        (scmds.continuousTone == SEC_TONE_ON ? 1 : 0) |
        (scmds.voltage==SEC_VOLTAGE_18 ? 2 : 0);

      scmds.miniCommand = SEC_MINI_NONE;
      scmds.numCommands=1;
      scmds.commands=&scmd;

      freq = qpsk.iFrequency;

      if (ioctl(fd_sec, SEC_SEND_SEQUENCE, &scmds) < 0)
       {  logPrintf (0, "setfront sec error (%s)\n", strerror (errno));
          close (fd_sec); close (fd_frontend); return (-1); }
   
      if (ioctl(fd_frontend, QPSK_TUNE, &qpsk) <0)
       { logPrintf (0, "setfront qpskfe error (%s)\n", strerror (errno));
          close (fd_sec); close (fd_frontend); return (-1); }
      if ((res = ioctl(fd_frontend, QPSK_GET_EVENT, &event))==-EBUFFEROVERFLOW)
            res = ioctl(fd_frontend, QPSK_GET_EVENT, &event);
      if (res < 0)
       { logPrintf (0, "get event qpskfe error\n", strerror (errno));
          close (fd_sec); close (fd_frontend); return (-1); }

      if (event.type != FE_COMPLETION_EV)
       { logPrintf (0, "could not tune to %d/%c/%d\n", Transponder->Frequency,
            Transponder->Polarisation, Transponder->SymbolRate);
          close (fd_sec); close (fd_frontend); return (-1); }

      newFrequency = Transponder->Frequency +
                        (event.u.completionEvent.iFrequency/1000 - freq);
      if (newFrequency != Transponder->Frequency)
         logPrintf (4, "frequency adjusted from %ld to %ld\n",
             Transponder->Frequency, newFrequency);

      close (fd_sec);
      close (fd_frontend);
   }

   return (newFrequency);
}


int processDvbCommand (struct OutputDevice *OutputDevice,
                      struct Command *Command)
{
   struct Channel         *Channel, *chan, *tchan;
   struct Command         *NewCommand;
   struct VtxPage         *VtxPage;
   time_t                  ThisTime;
   struct tm              *StartTime;
   char                   *sptr, *tptr;
   int                     newFreq, nextPage, curPage, curSub, i;

#define ImplicitOutputID OutputDevice->Identifier

   logPrintf (7, "%s, %d, %d - %d\n", xName (OutputDevice),
            OutputDevice->Identifier, Command->OutputID, Command->Code);

   switch (Command->Code)
   {
      case COMCODE_ON:
         if (openDvbDevice (OutputDevice) < 0)
          { logPrintf (0, "unable to open device %s\n",
               xName (OutputDevice)); break; }
         logPrintf (0, "Device %s switched on\n", xName (OutputDevice));
         OutputDevice->State = OUTST_IDLE;
         ScheduleDirect (COMCODE_TRANSPONDER_SCAN);
      break;

      case COMCODE_OFF:
         PurgeAllCommands (Command);
         CloseDependentInputDevices (IMPL_ALL);
         closeDvbDevice (OutputDevice);
         logPrintf (0, "Device %s switched off\n", xName (OutputDevice));
         OutputDevice->State = OUTST_OFF;
         ScheduleDirect (COMCODE_QUIT_PROGRAM);
      break;

      case COMCODE_TRANSPONDER_SCAN:
      {
         struct InputDevice *InpDev;
         struct Implication *Impl;
         struct Service     *Service;
         struct Program     *Program;
         struct Pid         *Pid;
         struct PidInfo     *PidInfo;
         struct Decriptor   *Descriptor;
         struct MEM_CHUNK  **OriginalKey;
         struct Satellite   *Satellite;

         CloseDependentInputDevices (IMPL_ALL);
         PurgeAllCommands (Command);
         if (!OutputDevice->InfoTransponder)
         {
            if (AvailableTransponders)
            {
               OutputDevice->InfoTransponder = (struct Transponder *)
                  AvailableTransponders->Head;
               NumChannels = 0;
               if (AvailableSatellites && xSize (AvailableSatellites))
               {
                  fprintf (outFile, "define_available_satellites {\n");
                  xForeach (AvailableSatellites, Satellite)
                     fprintf (outFile, "   {{%s} %d %d %d %d}\n",
                                 xName (Satellite), Satellite->DiseqcNr,
                                 Satellite->LnbLsof, Satellite->LnbFreqLow,
                                 Satellite->LnbFreqHigh);
                  fprintf (outFile, "}\n\n");
               }
               fprintf (outFile, "define_available_channels {\n");
            }
            else logPrintf (0, "no transponders defined to scan\n");
         }
         else
         {
            /*
             * analysis of scanned SDT/PAT/PMT
             */
            xForeach (InputDevices, InpDev)
            {
               xForeach (InpDev->Implications, Impl)
                  if ((Impl->Type & IMPL_DEPENDENT) &&
                      (Impl->Type & IMPL_SCAN) &&
                      (InpDev->State == INPDEV_FINISHED) &&
                      (Impl->OutputID == ImplicitOutputID)) break;
               if (Impl) break;
            }

            if (InpDev)
            {
               OriginalKey = xGetRemember ();
               xSetRemember (&TmpChanListRemember);
               if (!TmpChanList) TmpChanList = xNewList (NULL);

               xForeach (InpDev->ServiceList, Service)
               {
                  sprintf (sbuf, "channel%d", NumChannels++);
                  CreateChannel (Channel, 0, 0, sbuf, NULL, NULL,
                     Service->TransportStreamID,
                     OutputDevice->InfoTransponder->Frequency,
                     OutputDevice->InfoTransponder->Polarisation,
                     OutputDevice->InfoTransponder->SymbolRate,
                     Service->ServiceID, 0, 0, "unk", "auto", NULL);
                  Channel->Satellite = OutputDevice->InfoTransponder->Satellite;

                  if (GetConditionalAccess (Service->Status))
                     Channel->Camd = xSetText ("CA");
                  else Channel->Camd = xSetText ("FTA");

                  xForeach (Service->Descriptors, Descriptor)
                  {
                     switch (DescriptorTag (Descriptor))
                     {
                        case DESCR_SERVICE:
                           if (strlen (xName (Descriptor)))
                              xSetName (Channel, xName (Descriptor));
                           Channel->Provider = xSetText (
                              ((struct ServiceDescriptor *)
                                      Descriptor)->ServiceProvider);
                           switch (((struct ServiceDescriptor *)
                                      Descriptor)->ServiceType)
                           {
                              case 0x01:    /* TV */
                                 Channel->Type = xSetText ("TV");
                              break;

                              case 0x02:    /* Radio */
                                 Channel->Type = xSetText ("Radio");
                              break;

                              case 0x04:    /* NVOD */
                                 Channel->Type = xSetText ("NVOD");
                              break;
                           }
                        break;
                     }
                  }
                  xForeach (InpDev->ProgramList, Program)
                     if (Program->ProgramID == Service->ServiceID)
                        break;
                  if (Program)
                  {
                     xForeach (Program->Pids, Pid)
                     {
                        Channel->PcrPid = Pid->PcrPID;

                        xForeach (Pid->InfoList, PidInfo)
                        {
                           switch (PidInfo->StreamType)
                           {
                              case STREAMTYPE_ISO_VIDEO:
                              case STREAMTYPE_13818_VIDEO:
                              case STREAMTYPE_13522_MPEG:
                                 if (!Channel->VideoPid) Channel->VideoPid =
                                    PidInfo->ElementaryPid;
                              break;

                              case STREAMTYPE_11172_AUDIO:
                              case STREAMTYPE_13818_AUDIO:
                                 if (!Channel->AudioPid)
                                 {
                                    Channel->AudioPid = PidInfo->ElementaryPid;
                                    xForeach (PidInfo->Descriptors, Descriptor)
                                    {
                                       switch (DescriptorTag (Descriptor))
                                       {
                                          case DESCR_ISO_639_LANGUAGE:
                                             if (((struct
                                               Iso639LanguageDescriptor *)
                                               Descriptor)->LanguageCode[0])
                                                Channel->Language =
                                                   xSetText (((struct
                                                   Iso639LanguageDescriptor *)
                                                   Descriptor)->LanguageCode);
                                          break;
                                       }
                                    }
                                 }
                              break;

                              case STREAMTYPE_VIDEOTEXT:
                                 if (!Channel->VtxPid || (Channel->VtxPid !=
                                       PidInfo->ElementaryPid))
                                    Channel->VtxPid = PidInfo->ElementaryPid;
                              case STREAMTYPE_ITU_222:
                              case STREAMTYPE_13818_A:
                              case STREAMTYPE_13818_B:
                              case STREAMTYPE_13818_C:
                              case STREAMTYPE_13818_D:
                              case STREAMTYPE_13818_AUX:
                              break;
                           }
                        }
                     }
                  }
                  if (Channel->Type)
                  {
                     xAddTail (TmpChanList, Channel);
                     fprintf (outFile, "   {");
                     fprintf (outFile, "{%s} ", xName (Channel));
                     fprintf (outFile, "{%s} ", Channel->Provider);
                     fprintf (outFile, "%s ", Channel->Type);
                     fprintf (outFile, "%d ", Channel->TransportStreamID);
                     fprintf (outFile, "%d ", Channel->Frequency/1000);
                     fprintf (outFile, "%c ", Channel->Polarisation);
                     fprintf (outFile, "%d ", Channel->SymbolRate);
                     fprintf (outFile, "%d ", Channel->ServiceID);
                     fprintf (outFile, "%d ", Channel->VideoPid);
                     fprintf (outFile, "%d ", Channel->AudioPid);
                     fprintf (outFile, "%d ", Channel->PcrPid);
                     fprintf (outFile, "%s ", Channel->Language);
                     fprintf (outFile, "%s ", Channel->Fec);
                     fprintf (outFile, "%s ", Channel->Camd);
                     if (Channel->Satellite)
                        fprintf (outFile, "{%s}", xName (Channel->Satellite));
                     fprintf (outFile, "}\n");
                     if (vdrFile)
                     {
                        for (i = 0; i < strlen (xName (Channel)); i++)
                           if (xName(Channel)[i] == ':') sbuf[i] = '|';
                           else sbuf[i] = xName(Channel)[i];
                        sbuf[i] = '\0'; fprintf (vdrFile, "%s:", sbuf);
                        fprintf (vdrFile, "%d:", Channel->Frequency/1000);
                        fprintf (vdrFile, "%c:", Channel->Polarisation);
                        if (Channel->Satellite)
                          fprintf (vdrFile,"%d:",Channel->Satellite->DiseqcNr);
                        else fprintf (vdrFile, "0:");
                        fprintf (vdrFile, "%d:", Channel->SymbolRate);
                        fprintf (vdrFile, "%d:", Channel->VideoPid);
                        fprintf (vdrFile, "%d:", Channel->AudioPid);
                        fprintf (vdrFile, "%d:", Channel->VtxPid);
                        if (!strcmp (Channel->Camd, "FTA"))
                           fprintf (vdrFile, "0:");
                        else fprintf (vdrFile, "1:");
                        fprintf (vdrFile, "%d\n", Channel->ServiceID);
                     }
                  }
               }
               xSetRemember (OriginalKey);
            }

            OutputDevice->InfoTransponder = (struct Transponder *)
                 xSucc (OutputDevice->InfoTransponder);
         }
         if (OutputDevice->InfoTransponder)
         {
            if ((newFreq = setDvbChannel (OutputDevice,
                   OutputDevice->InfoTransponder)) < 0)
            {
               logPrintf (0, "could not tune transponder %d/%c/%d\n",
                  OutputDevice->InfoTransponder->Frequency,
                  OutputDevice->InfoTransponder->Polarisation,
                  OutputDevice->InfoTransponder->SymbolRate);
               ScheduleDirect (COMCODE_TRANSPONDER_SCAN);
            }
            else
            {
               logPrintf (0, "scanning %d/%c/%d\n",
                  OutputDevice->InfoTransponder->Frequency,
                  OutputDevice->InfoTransponder->Polarisation,
                  OutputDevice->InfoTransponder->SymbolRate);
               OpenDependentInputDevices (IMPL_SCAN);
               ScheduleDelayed (COMCODE_TRANSPONDER_SCAN,
                  GetGlobalOption ("scanTimeout")*10);
            }
         }
         else
         {
            /* last scan was done */
            fprintf (outFile, "}\n");
            ScheduleDirect (COMCODE_OFF);
         }
      }
      break;
   }

#undef ImplicitOutputID

   xRemove (CommandQueue, Command);
   if (Command->Code == COMCODE_QUIT_PROGRAM) return (1);
   return (0);
}

