//////////////////////////////////////////////////////////////
///                                                        ///
/// dvb_si.c: routines to parse the DVB-SI stream          ///
///                                                        ///
//////////////////////////////////////////////////////////////

// $Revision: 1.2 $
// $Date: 2001/06/23 16:33:40 $
// $Author: hakenes $
//
//   (C) 2001 Rolf Hakenes <hakenes@hippomi.de>, under the GNU GPL.
//
// dtv_scan is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2, or (at your option)
// any later version.
//
// dtv_scan is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You may have received a copy of the GNU General Public License
// along with dtv_scan; see the file COPYING.  If not, write to the
// Free Software Foundation, Inc., 59 Temple Place - Suite 330,
// Boston, MA 02111-1307, USA.


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <fcntl.h>

#include <ost/dmx.h>

#include <libsi.h>
#include <si_tables.h>

#include "definitions.h"
#include "structures.h"
#include "prototypes.h"



static u_char siBuffer[MAX_SECTION_BUFFER];  /* maximum section */


int openScan(struct InputDevice *InputDevice)
{
   SetupFilterAndSection (InputDevice, PID_SDT, TID_SDT_ACT, 0);

   InputDevice->State = INPDEV_GET_SDT;
   if (InputDevice->RememberKey) 
   {
      xMemFreeAll (&InputDevice->RememberKey);
      InputDevice->ProgramList = NULL;
      InputDevice->Program = NULL;
      InputDevice->ServiceList = NULL;
      InputDevice->Service = NULL;
      InputDevice->EventList = NULL;
      InputDevice->Event = NULL;
   }
   return (InputDevice->FileHandle);
}


void closeSi(struct InputDevice *InputDevice)
{
   if (InputDevice->FileHandle)
   {
      ioctl (InputDevice->FileHandle, DMX_STOP, 0);
      close (InputDevice->FileHandle);
      InputDevice->FileHandle = 0;
      InputDevice->Status = INPDEV_DISABLED;
   }
   return;
}


int getScanEvent (struct InputDevice *InputDevice)
{
   struct OutputDevice                   *OutputDevice;
   struct Service                        *Service;
   struct Descriptor                     *Descriptor;
   struct LIST                           *Result;
   struct MEM_CHUNK                     **OriginalKey;
   unsigned short                         flag, EventListSize;
   int                                    count;
        
   if ((count = read (InputDevice->FileHandle, siBuffer, MAX_SECTION_BUFFER)) < 0)
      return (0);

#define ImplicitOutputID OutputDevice->Identifier

   FindOutputDevice (InputDevice, OutputDevice);

   if (OutputDevice)
   {
      OriginalKey = xGetRemember ();
      xSetRemember (&InputDevice->RememberKey);
      switch (InputDevice->State)
      {
         case INPDEV_GET_SDT:
         logPrintf (4, "InpDev Scan GetSDT\n");
         {
            Result = siParseSDT (siBuffer);
            xConCat (InputDevice->ServiceList, Result);
            if (Result && TableHasMoreSections (siBuffer))
               SetupFilterAndSection (InputDevice, PID_SDT,
                   TID_SDT_ACT, GetSectionNumber(siBuffer)+1);
            else
            {
               logPrintf (2, "SDT acquisition finished\n");
               SetupFilterAndSection (InputDevice, PID_PAT, TID_PAT, 0);
               InputDevice->State = INPDEV_GET_PAT;
            }
         }
         break;

         case INPDEV_GET_PAT:
         logPrintf (4, "InpDev Scan GetPAT\n");
         {
            Result = siParsePAT (siBuffer);
            xConCat (InputDevice->ProgramList, Result);
            if (Result && TableHasMoreSections (siBuffer))
               SetupFilterAndSection (InputDevice, PID_PAT,
                   TID_PAT, GetSectionNumber(siBuffer)+1);
            else
            {
               logPrintf (2, "PAT acquisition finished, %d programs to scan\n",
                            xSize (InputDevice->ProgramList));
               InputDevice->Program = (struct Program *)
                                            InputDevice->ProgramList->Head;
               while (InputDevice->Program && !isValidProgram (InputDevice))
                  InputDevice->Program = (struct Program *)
                                            xSucc (InputDevice->Program);
               if (!InputDevice->Program)
               {
                  logPrintf (2, "No Programs found\n");
                  ScheduleDirect (COMCODE_TRANSPONDER_SCAN);
                  InputDevice->State = INPDEV_ERROR;
                  closeSi (InputDevice);
               }
               else
               {
                  logPrintf (4, "Setup PMT Filter on %d/%d\n",
                     InputDevice->Program->NetworkPID,
                     InputDevice->Program->ProgramID);
                  SetupFilterOnID (InputDevice,
                     InputDevice->Program->NetworkPID, TID_PMT,
                     0, InputDevice->Program->ProgramID);
                  InputDevice->State = INPDEV_GET_TV_PMT;
               }
            }
         }
         break;

         case INPDEV_GET_TV_PMT:
         logPrintf (4, "InpDev Scan GetTVPMT\n");
         {
            struct Pid       *Pid;

            if (Pid = siParsePMT (siBuffer))
               xAddTail (InputDevice->Program->Pids, Pid);

            InputDevice->Program = (struct Program *)
                                         xSucc (InputDevice->Program);
            while (InputDevice->Program && !isValidProgram (InputDevice))
               InputDevice->Program = (struct Program *)
                                         xSucc (InputDevice->Program);
            if (!InputDevice->Program)
            {
               logPrintf (2, "PMT acquisition finished\n");
               ScheduleDirect (COMCODE_TRANSPONDER_SCAN);
               InputDevice->State = INPDEV_FINISHED;
               closeSi (InputDevice);
            }
            else
            {
               logPrintf (4, "Setup PMT Filter on %d/%d\n",
                  InputDevice->Program->NetworkPID,
                  InputDevice->Program->ProgramID);
               SetupFilterOnID (InputDevice,
                  InputDevice->Program->NetworkPID, TID_PMT,
                  0, InputDevice->Program->ProgramID);
               InputDevice->State = INPDEV_GET_TV_PMT;
            }
         }
         break;
      }
      xSetRemember (OriginalKey);
   }

#undef ImplicitOutputID

   return (0);
}

int isValidProgram (struct InputDevice *InputDevice)
{
   struct Service                        *Service;
   struct Descriptor                     *Descriptor;

   if (!InputDevice->Program ||
       !InputDevice->Program->ProgramID) return (0);

   InputDevice->Service = NULL;

   xForeach (InputDevice->ServiceList, Service)
      if (Service->ServiceID == InputDevice->Program->ProgramID)
         xForeach (Service->Descriptors, Descriptor)
         {
            switch (DescriptorTag (Descriptor))
            {
               case DESCR_SERVICE:
                  switch (((struct ServiceDescriptor *)
                             Descriptor)->ServiceType)
                  {
                     case 0x01:    /* TV */
                        logPrintf(4,"Service %d is TV (%s)\n",
                          Service->ServiceID,xName (Descriptor));
                        InputDevice->Service = Service; return (1);
                     break;
                     case 0x02:    /* Radio */
                        logPrintf(4,"Service %d is Radio (%s)\n",
                          Service->ServiceID,xName (Descriptor));
                        InputDevice->Service = Service; return (1);
                     break;
                     case 0x04:    /* NVOD */
                        logPrintf(4,"Service %d is NVOD (%s)\n",
                          Service->ServiceID,xName (Descriptor));
                        InputDevice->Service = Service; return (1);
                     break;
                  }
               break;
            }
         }
   return (0);
}
