/*
 * dcl.h --
 *
 *	This header file describes the externally-visible facilities
 *	of the Dcl interpreter.
 *
 * Copyright (c) 1987-1994 The Regents of the University of California.
 * Copyright (c) 1994-1995 Sun Microsystems, Inc.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * @(#) dcl.h 1.153 95/06/27 15:42:31
 */

#ifndef _DCL
#define _DCL

#ifndef BUFSIZ
#include <stdio.h>
#endif

#define DCL_VERSION "7.4"
#define DCL_MAJOR_VERSION 7
#define DCL_MINOR_VERSION 4

/*
 * Definitions that allow this header file to be used either with or
 * without ANSI C features like function prototypes.
 */

#undef _ANSI_ARGS_
#undef CONST
#define _USING_PROTOTYPES_ 1
#define _ANSI_ARGS_(x)	x
#define CONST const
#define VARARGS(first) (first, ...)

#ifdef __cplusplus
#   define EXTERN extern "C"
#else
#   define EXTERN extern
#endif

/*
 * Macro to use instead of "void" for arguments that must have
 * type "void *" in ANSI C;  maps them to type "char *" in
 * non-ANSI systems.
 */

#ifndef VOID
#   ifdef __STDC__
#       define VOID void
#   else
#       define VOID char
#   endif
#endif

/*
 * Miscellaneous declarations (to allow Dcl to be used stand-alone,
 * without the rest of Sprite).
 */

#ifndef NULL
#define NULL 0
#endif

#ifndef _CLIENTDATA
#   if defined(__STDC__) || defined(__cplusplus)
    typedef void *ClientData;
#   else
    typedef int *ClientData;
#   endif /* __STDC__ */
#define _CLIENTDATA
#endif

/*
 * Data structures defined opaquely in this module.  The definitions
 * below just provide dummy types.  A few fields are made visible in
 * Dcl_Interp structures, namely those for returning string values.
 * Note:  any change to the Dcl_Interp definition below must be mirrored
 * in the "real" definition in dclInt.h.
 */

typedef struct Dcl_Interp{
    char *result;		/* Points to result string returned by last
				 * command. */
    void (*freeProc) _ANSI_ARGS_((char *blockPtr));
				/* Zero means result is statically allocated.
				 * If non-zero, gives address of procedure
				 * to invoke to free the result.  Must be
				 * freed by Dcl_Eval before executing next
				 * command. */
    int errorLine;		/* When DCL_ERROR is returned, this gives
				 * the line number within the command where
				 * the error occurred (1 means first line). */
} Dcl_Interp;

typedef int *Dcl_Trace;
typedef int *Dcl_Command;
typedef struct Dcl_RegExp_ *Dcl_RegExp;

/*
 * When a DCL command returns, the string pointer interp->result points to
 * a string containing return information from the command.  In addition,
 * the command procedure returns an integer value, which is one of the
 * following:
 *
 * DCL_OK		Command completed normally;  interp->result contains
 *			the command's result.
 * DCL_ERROR		The command couldn't be completed successfully;
 *			interp->result describes what went wrong.
 * DCL_RETURN		The command requests that the current procedure
 *			return;  interp->result contains the procedure's
 *			return value.
 * DCL_BREAK		The command requests that the innermost loop
 *			be exited;  interp->result is meaningless.
 * DCL_CONTINUE		Go on to the next iteration of the current loop;
 *			interp->result is meaningless.
 */

#define DCL_OK		0
#define DCL_ERROR	1
#define DCL_RETURN	2
#define DCL_BREAK	3
#define DCL_CONTINUE	4

#define DCL_RESULT_SIZE 200

/*
 * Argument descriptors for math function callbacks in expressions:
 */

typedef enum {DCL_INT, DCL_DOUBLE, DCL_EITHER} Dcl_ValueType;
typedef struct Dcl_Value {
    Dcl_ValueType type;		/* Indicates intValue or doubleValue is
				 * valid, or both. */
    long intValue;		/* Integer value. */
    double doubleValue;		/* Double-precision floating value. */
} Dcl_Value;

/*
 * Procedure types defined by Dcl:
 */

typedef int (Dcl_AppInitProc) _ANSI_ARGS_((Dcl_Interp *interp));
typedef void (Dcl_CmdDeleteProc) _ANSI_ARGS_((ClientData clientData));
typedef int (Dcl_CmdProc) _ANSI_ARGS_((ClientData clientData,
	Dcl_Interp *interp, int argc, char *argv[]));
typedef void (Dcl_CmdTraceProc) _ANSI_ARGS_((ClientData clientData,
	Dcl_Interp *interp, int level, char *command, Dcl_CmdProc *proc,
	ClientData cmdClientData, int argc, char *argv[]));
typedef void (Dcl_FreeProc) _ANSI_ARGS_((char *blockPtr));
typedef void (Dcl_InterpDeleteProc) _ANSI_ARGS_((ClientData clientData,
	Dcl_Interp *interp));
typedef int (Dcl_MathProc) _ANSI_ARGS_((ClientData clientData,
	Dcl_Interp *interp, Dcl_Value *args, Dcl_Value *resultPtr));
typedef char *(Dcl_VarTraceProc) _ANSI_ARGS_((ClientData clientData,
	Dcl_Interp *interp, char *part1, char *part2, int flags));

/*
 * The structure returned by Dcl_GetCmdInfo and passed into
 * Dcl_SetCmdInfo:
 */

typedef struct Dcl_CmdInfo {
    Dcl_CmdProc *proc;			/* Procedure that implements command. */
    ClientData clientData;		/* ClientData passed to proc. */
    Dcl_CmdDeleteProc *deleteProc;	/* Procedure to call when command
					 * is deleted. */
    ClientData deleteData;		/* Value to pass to deleteProc (usually
					 * the same as clientData). */
} Dcl_CmdInfo;

/*
 * The structure defined below is used to hold dynamic strings.  The only
 * field that clients should use is the string field, and they should
 * never modify it.
 */

#define DCL_DSTRING_STATIC_SIZE 200
typedef struct Dcl_DString {
    char *string;		/* Points to beginning of string:  either
				 * staticSpace below or a malloc'ed array. */
    int length;			/* Number of non-NULL characters in the
				 * string. */
    int spaceAvl;		/* Total number of bytes available for the
				 * string and its terminating NULL char. */
    char staticSpace[DCL_DSTRING_STATIC_SIZE];
				/* Space to use in common case where string
				 * is small. */
} Dcl_DString;

#define Dcl_DStringLength(dsPtr) ((dsPtr)->length)
#define Dcl_DStringValue(dsPtr) ((dsPtr)->string)
#define Dcl_DStringTrunc Dcl_DStringSetLength

/*
 * Definitions for the maximum number of digits of precision that may
 * be specified in the "dcl_precision" variable, and the number of
 * characters of buffer space required by Dcl_PrintDouble.
 */

#define DCL_MAX_PREC 17
#define DCL_DOUBLE_SPACE (DCL_MAX_PREC+10)

/*
 * Flag that may be passed to Dcl_ConvertElement to force it not to
 * output braces (careful!  if you change this flag be sure to change
 * the definitions at the front of dclUtil.c).
 */

#define DCL_DONT_USE_BRACES	1

/*
 * Flag values passed to Dcl_RecordAndEval.
 * WARNING: these bit choices must not conflict with the bit choices
 * for evalFlag bits in dclInt.h!!
 */

#define DCL_NO_EVAL		0x10000
#define DCL_EVAL_GLOBAL		0x20000

/*
 * Special freeProc values that may be passed to Dcl_SetResult (see
 * the man page for details):
 */

#define DCL_VOLATILE	((Dcl_FreeProc *) 1)
#define DCL_STATIC	((Dcl_FreeProc *) 0)
#define DCL_DYNAMIC	((Dcl_FreeProc *) 3)

/*
 * Flag values passed to variable-related procedures.
 */

#define DCL_GLOBAL_ONLY		1
#define DCL_APPEND_VALUE	2
#define DCL_LIST_ELEMENT	4
#define DCL_TRACE_READS		0x10
#define DCL_TRACE_WRITES	0x20
#define DCL_TRACE_UNSETS	0x40
#define DCL_TRACE_DESTROYED	0x80
#define DCL_INTERP_DESTROYED	0x100
#define DCL_LEAVE_ERR_MSG	0x200

/*
 * Types for linked variables:
 */

#define DCL_LINK_INT		1
#define DCL_LINK_DOUBLE		2
#define DCL_LINK_BOOLEAN	3
#define DCL_LINK_STRING		4
#define DCL_LINK_READ_ONLY	0x80

/*
 * Permission flags for files:
 */

#define DCL_FILE_READABLE	1
#define DCL_FILE_WRITABLE	2

/*
 * The following declarations either map ckalloc and ckfree to
 * malloc and free, or they map them to procedures with all sorts
 * of debugging hooks defined in dclCkalloc.c.
 */

#ifdef DCL_MEM_DEBUG

EXTERN char *		Dcl_DbCkalloc _ANSI_ARGS_((unsigned int size,
			    char *file, int line));
EXTERN int		Dcl_DbCkfree _ANSI_ARGS_((char *ptr,
			    char *file, int line));
EXTERN char *		Dcl_DbCkrealloc _ANSI_ARGS_((char *ptr,
			    unsigned int size, char *file, int line));
EXTERN int		Dcl_DumpActiveMemory _ANSI_ARGS_((char *fileName));
EXTERN void		Dcl_ValidateAllMemory _ANSI_ARGS_((char *file,
			    int line));
#  define ckalloc(x) Dcl_DbCkalloc(x, __FILE__, __LINE__)
#  define ckfree(x)  Dcl_DbCkfree(x, __FILE__, __LINE__)
#  define ckrealloc(x,y) Dcl_DbCkrealloc((x), (y),__FILE__, __LINE__)

#else

#  define ckalloc(x) malloc(x)
#  define ckfree(x)  free(x)
#  define ckrealloc(x,y) realloc(x,y)
#  define Dcl_DumpActiveMemory(x)
#  define Dcl_ValidateAllMemory(x,y)

#endif /* DCL_MEM_DEBUG */

/*
 * Macro to free up result of interpreter.
 */

#define Dcl_FreeResult(interp)					\
    if ((interp)->freeProc != 0) {				\
	if ((interp)->freeProc == (Dcl_FreeProc *) free) {	\
	    ckfree((interp)->result);				\
	} else {						\
	    (*(interp)->freeProc)((interp)->result);		\
	}							\
	(interp)->freeProc = 0;					\
    }

/*
 * Forward declaration of Dcl_HashTable.  Needed by some C++ compilers
 * to prevent errors when the forward reference to Dcl_HashTable is
 * encountered in the Dcl_HashEntry structure.
 */

#ifdef __cplusplus
struct Dcl_HashTable;
#endif

/*
 * Structure definition for an entry in a hash table.  No-one outside
 * Dcl should access any of these fields directly;  use the macros
 * defined below.
 */

typedef struct Dcl_HashEntry {
    struct Dcl_HashEntry *nextPtr;	/* Pointer to next entry in this
					 * hash bucket, or NULL for end of
					 * chain. */
    struct Dcl_HashTable *tablePtr;	/* Pointer to table containing entry. */
    struct Dcl_HashEntry **bucketPtr;	/* Pointer to bucket that points to
					 * first entry in this entry's chain:
					 * used for deleting the entry. */
    ClientData clientData;		/* Application stores something here
					 * with Dcl_SetHashValue. */
    union {				/* Key has one of these forms: */
	char *oneWordValue;		/* One-word value for key. */
	int words[1];			/* Multiple integer words for key.
					 * The actual size will be as large
					 * as necessary for this table's
					 * keys. */
	char string[4];			/* String for key.  The actual size
					 * will be as large as needed to hold
					 * the key. */
    } key;				/* MUST BE LAST FIELD IN RECORD!! */
} Dcl_HashEntry;

/*
 * Structure definition for a hash table.  Must be in dcl.h so clients
 * can allocate space for these structures, but clients should never
 * access any fields in this structure.
 */

#define DCL_SMALL_HASH_TABLE 4
typedef struct Dcl_HashTable {
    Dcl_HashEntry **buckets;		/* Pointer to bucket array.  Each
					 * element points to first entry in
					 * bucket's hash chain, or NULL. */
    Dcl_HashEntry *staticBuckets[DCL_SMALL_HASH_TABLE];
					/* Bucket array used for small tables
					 * (to avoid mallocs and frees). */
    int numBuckets;			/* Total number of buckets allocated
					 * at **bucketPtr. */
    int numEntries;			/* Total number of entries present
					 * in table. */
    int rebuildSize;			/* Enlarge table when numEntries gets
					 * to be this large. */
    int downShift;			/* Shift count used in hashing
					 * function.  Designed to use high-
					 * order bits of randomized keys. */
    int mask;				/* Mask value used in hashing
					 * function. */
    int keyType;			/* Type of keys used in this table. 
					 * It's either DCL_STRING_KEYS,
					 * DCL_ONE_WORD_KEYS, or an integer
					 * giving the number of ints in a
					 */
    Dcl_HashEntry *(*findProc) _ANSI_ARGS_((struct Dcl_HashTable *tablePtr,
	    char *key));
    Dcl_HashEntry *(*createProc) _ANSI_ARGS_((struct Dcl_HashTable *tablePtr,
	    char *key, int *newPtr));
} Dcl_HashTable;

/*
 * Structure definition for information used to keep track of searches
 * through hash tables:
 */

typedef struct Dcl_HashSearch {
    Dcl_HashTable *tablePtr;		/* Table being searched. */
    int nextIndex;			/* Index of next bucket to be
					 * enumerated after present one. */
    Dcl_HashEntry *nextEntryPtr;	/* Next entry to be enumerated in the
					 * the current bucket. */
} Dcl_HashSearch;

/*
 * Acceptable key types for hash tables:
 */

#define DCL_STRING_KEYS		0
#define DCL_ONE_WORD_KEYS	1

/*
 * Macros for clients to use to access fields of hash entries:
 */

#define Dcl_GetHashValue(h) ((h)->clientData)
#define Dcl_SetHashValue(h, value) ((h)->clientData = (ClientData) (value))
#define Dcl_GetHashKey(tablePtr, h) \
    ((char *) (((tablePtr)->keyType == DCL_ONE_WORD_KEYS) ? (h)->key.oneWordValue \
						: (h)->key.string))

/*
 * Macros to use for clients to use to invoke find and create procedures
 * for hash tables:
 */

#define Dcl_FindHashEntry(tablePtr, key) \
	(*((tablePtr)->findProc))(tablePtr, key)
#define Dcl_CreateHashEntry(tablePtr, key, newPtr) \
	(*((tablePtr)->createProc))(tablePtr, key, newPtr)

/*
 * Exported Dcl variables:
 */

EXTERN void		(*dcl_FileCloseProc) _ANSI_ARGS_((FILE *f));
EXTERN char *		dcl_RcFileName;

/*
 * Exported Dcl procedures:
 */

EXTERN void		Dcl_AddErrorInfo _ANSI_ARGS_((Dcl_Interp *interp,
			    char *message));
EXTERN void		Dcl_AllowExceptions _ANSI_ARGS_((Dcl_Interp *interp));
EXTERN void		Dcl_AppendElement _ANSI_ARGS_((Dcl_Interp *interp,
			    char *string));
EXTERN void		Dcl_AppendResult _ANSI_ARGS_(VARARGS(Dcl_Interp *interp));
EXTERN int		Dcl_AppInit _ANSI_ARGS_((Dcl_Interp *interp));
EXTERN char		Dcl_Backslash _ANSI_ARGS_((char *src,
			    int *readPtr));
EXTERN void		Dcl_CallWhenDeleted _ANSI_ARGS_((Dcl_Interp *interp,
			    Dcl_InterpDeleteProc *proc,
			    ClientData clientData));
EXTERN int		Dcl_CommandComplete _ANSI_ARGS_((char *cmd));
EXTERN char *		Dcl_Concat _ANSI_ARGS_((int argc, char **argv));
EXTERN int		Dcl_ConvertElement _ANSI_ARGS_((char *src,
			    char *dst, int flags));
EXTERN Dcl_Command	Dcl_CreateCommand _ANSI_ARGS_((Dcl_Interp *interp,
			    char *cmdName, Dcl_CmdProc *proc,
			    ClientData clientData,
			    Dcl_CmdDeleteProc *deleteProc));
EXTERN Dcl_Interp *	Dcl_CreateInterp _ANSI_ARGS_((void));
EXTERN void		Dcl_CreateMathFunc _ANSI_ARGS_((Dcl_Interp *interp,
			    char *name, int numArgs, Dcl_ValueType *argTypes,
			    Dcl_MathProc *proc, ClientData clientData));
EXTERN int		Dcl_CreatePipeline _ANSI_ARGS_((Dcl_Interp *interp,
			    int argc, char **argv, int **pidArrayPtr,
			    int *inPipePtr, int *outPipePtr,
			    int *errFilePtr));
EXTERN Dcl_Trace	Dcl_CreateTrace _ANSI_ARGS_((Dcl_Interp *interp,
			    int level, Dcl_CmdTraceProc *proc,
			    ClientData clientData));
EXTERN int		Dcl_DeleteCommand _ANSI_ARGS_((Dcl_Interp *interp,
			    char *cmdName));
EXTERN void		Dcl_DeleteHashEntry _ANSI_ARGS_((
			    Dcl_HashEntry *entryPtr));
EXTERN void		Dcl_DeleteHashTable _ANSI_ARGS_((
			    Dcl_HashTable *tablePtr));
EXTERN void		Dcl_DeleteInterp _ANSI_ARGS_((Dcl_Interp *interp));
EXTERN void		Dcl_DeleteTrace _ANSI_ARGS_((Dcl_Interp *interp,
			    Dcl_Trace trace));
EXTERN void		Dcl_DetachPids _ANSI_ARGS_((int numPids, int *pidPtr));
EXTERN void		Dcl_DontCallWhenDeleted _ANSI_ARGS_((
			    Dcl_Interp *interp, Dcl_InterpDeleteProc *proc,
			    ClientData clientData));
EXTERN char *		Dcl_DStringAppend _ANSI_ARGS_((Dcl_DString *dsPtr,
			    char *string, int length));
EXTERN char *		Dcl_DStringAppendElement _ANSI_ARGS_((
			    Dcl_DString *dsPtr, char *string));
EXTERN void		Dcl_DStringEndSublist _ANSI_ARGS_((Dcl_DString *dsPtr));
EXTERN void		Dcl_DStringFree _ANSI_ARGS_((Dcl_DString *dsPtr));
EXTERN void		Dcl_DStringGetResult _ANSI_ARGS_((Dcl_Interp *interp,
			    Dcl_DString *dsPtr));
EXTERN void		Dcl_DStringInit _ANSI_ARGS_((Dcl_DString *dsPtr));
EXTERN void		Dcl_DStringResult _ANSI_ARGS_((Dcl_Interp *interp,
			    Dcl_DString *dsPtr));
EXTERN void		Dcl_DStringSetLength _ANSI_ARGS_((Dcl_DString *dsPtr,
			    int length));
EXTERN void		Dcl_DStringStartSublist _ANSI_ARGS_((
			    Dcl_DString *dsPtr));
EXTERN void		Dcl_EnterFile _ANSI_ARGS_((Dcl_Interp *interp,
			    FILE *file, int permissions));
EXTERN char *		Dcl_ErrnoId _ANSI_ARGS_((void));
EXTERN int		Dcl_Eval _ANSI_ARGS_((Dcl_Interp *interp, char *cmd));
EXTERN int		Dcl_EvalFile _ANSI_ARGS_((Dcl_Interp *interp,
			    char *fileName));
EXTERN int		Dcl_ExprBoolean _ANSI_ARGS_((Dcl_Interp *interp,
			    char *string, int *ptr));
EXTERN int		Dcl_ExprDouble _ANSI_ARGS_((Dcl_Interp *interp,
			    char *string, double *ptr));
EXTERN int		Dcl_ExprLong _ANSI_ARGS_((Dcl_Interp *interp,
			    char *string, long *ptr));
EXTERN int		Dcl_ExprString _ANSI_ARGS_((Dcl_Interp *interp,
			    char *string));
EXTERN int		Dcl_FilePermissions _ANSI_ARGS_((FILE *file));
EXTERN Dcl_HashEntry *	Dcl_FirstHashEntry _ANSI_ARGS_((
			    Dcl_HashTable *tablePtr,
			    Dcl_HashSearch *searchPtr));
EXTERN int		Dcl_GetBoolean _ANSI_ARGS_((Dcl_Interp *interp,
			    char *string, int *boolPtr));
EXTERN int		Dcl_GetCommandInfo _ANSI_ARGS_((Dcl_Interp *interp,
			    char *cmdName, Dcl_CmdInfo *infoPtr));
EXTERN char *		Dcl_GetCommandName _ANSI_ARGS_((Dcl_Interp *interp,
			    Dcl_Command command));
EXTERN int		Dcl_GetDouble _ANSI_ARGS_((Dcl_Interp *interp,
			    char *string, double *doublePtr));
EXTERN int		Dcl_GetInt _ANSI_ARGS_((Dcl_Interp *interp,
			    char *string, int *intPtr));
EXTERN int		Dcl_GetOpenFile _ANSI_ARGS_((Dcl_Interp *interp,
			    char *string, int write, int checkUsage,
			    FILE **filePtr));
EXTERN char *		Dcl_GetVar _ANSI_ARGS_((Dcl_Interp *interp,
			    char *varName, int flags));
EXTERN char *		Dcl_GetVar2 _ANSI_ARGS_((Dcl_Interp *interp,
			    char *part1, char *part2, int flags));
EXTERN int		Dcl_GlobalEval _ANSI_ARGS_((Dcl_Interp *interp,
			    char *command));
EXTERN char *		Dcl_HashStats _ANSI_ARGS_((Dcl_HashTable *tablePtr));
EXTERN int		Dcl_Init _ANSI_ARGS_((Dcl_Interp *interp));
EXTERN void		Dcl_InitHashTable _ANSI_ARGS_((Dcl_HashTable *tablePtr,
			    int keyType));
EXTERN void		Dcl_InitMemory _ANSI_ARGS_((Dcl_Interp *interp));
EXTERN int		Dcl_LinkVar _ANSI_ARGS_((Dcl_Interp *interp,
			    char *varName, char *addr, int type));
EXTERN void		Dcl_Main _ANSI_ARGS_((int argc, char **argv,
			    Dcl_AppInitProc *appInitProc));
EXTERN char *		Dcl_Merge _ANSI_ARGS_((int argc, char **argv));
EXTERN Dcl_HashEntry *	Dcl_NextHashEntry _ANSI_ARGS_((
			    Dcl_HashSearch *searchPtr));
EXTERN char *		Dcl_ParseVar _ANSI_ARGS_((Dcl_Interp *interp,
			    char *string, char **termPtr));
EXTERN char *		Dcl_PosixError _ANSI_ARGS_((Dcl_Interp *interp));
EXTERN void		Dcl_PrintDouble _ANSI_ARGS_((Dcl_Interp *interp,
			    double value, char *dst));
EXTERN void		Dcl_ReapDetachedProcs _ANSI_ARGS_((void));
EXTERN int		Dcl_RecordAndEval _ANSI_ARGS_((Dcl_Interp *interp,
			    char *cmd, int flags));
EXTERN Dcl_RegExp	Dcl_RegExpCompile _ANSI_ARGS_((Dcl_Interp *interp,
			    char *string));
EXTERN int		Dcl_RegExpExec _ANSI_ARGS_((Dcl_Interp *interp,
			    Dcl_RegExp regexp, char *string, char *start));
EXTERN int		Dcl_RegExpMatch _ANSI_ARGS_((Dcl_Interp *interp,
			    char *string, char *pattern));
EXTERN void		Dcl_RegExpRange _ANSI_ARGS_((Dcl_RegExp regexp,
			    int index, char **startPtr, char **endPtr));
EXTERN void		Dcl_ResetResult _ANSI_ARGS_((Dcl_Interp *interp));
#define Dcl_Return Dcl_SetResult
EXTERN int		Dcl_ScanElement _ANSI_ARGS_((char *string,
			    int *flagPtr));
EXTERN int		Dcl_SetCommandInfo _ANSI_ARGS_((Dcl_Interp *interp,
			    char *cmdName, Dcl_CmdInfo *infoPtr));
EXTERN void		Dcl_SetErrorCode _ANSI_ARGS_(
			    VARARGS(Dcl_Interp *interp));
EXTERN int		Dcl_SetRecursionLimit _ANSI_ARGS_((Dcl_Interp *interp,
			    int depth));
EXTERN void		Dcl_SetResult _ANSI_ARGS_((Dcl_Interp *interp,
			    char *string, Dcl_FreeProc *freeProc));
EXTERN char *		Dcl_SetVar _ANSI_ARGS_((Dcl_Interp *interp,
			    char *varName, char *newValue, int flags));
EXTERN char *		Dcl_SetVar2 _ANSI_ARGS_((Dcl_Interp *interp,
			    char *part1, char *part2, char *newValue,
			    int flags));
EXTERN char *		Dcl_SignalId _ANSI_ARGS_((int sig));
EXTERN char *		Dcl_SignalMsg _ANSI_ARGS_((int sig));
EXTERN int		Dcl_SplitList _ANSI_ARGS_((Dcl_Interp *interp,
			    char *list, int *argcPtr, char ***argvPtr));
EXTERN int		Dcl_StringMatch _ANSI_ARGS_((char *string,
			    char *pattern));
EXTERN char *		Dcl_TildeSubst _ANSI_ARGS_((Dcl_Interp *interp,
			    char *name, Dcl_DString *bufferPtr));
EXTERN int		Dcl_TraceVar _ANSI_ARGS_((Dcl_Interp *interp,
			    char *varName, int flags, Dcl_VarTraceProc *proc,
			    ClientData clientData));
EXTERN int		Dcl_TraceVar2 _ANSI_ARGS_((Dcl_Interp *interp,
			    char *part1, char *part2, int flags,
			    Dcl_VarTraceProc *proc, ClientData clientData));
EXTERN void		Dcl_UnlinkVar _ANSI_ARGS_((Dcl_Interp *interp,
			    char *varName));
EXTERN int		Dcl_UnsetVar _ANSI_ARGS_((Dcl_Interp *interp,
			    char *varName, int flags));
EXTERN int		Dcl_UnsetVar2 _ANSI_ARGS_((Dcl_Interp *interp,
			    char *part1, char *part2, int flags));
EXTERN void		Dcl_UntraceVar _ANSI_ARGS_((Dcl_Interp *interp,
			    char *varName, int flags, Dcl_VarTraceProc *proc,
			    ClientData clientData));
EXTERN void		Dcl_UntraceVar2 _ANSI_ARGS_((Dcl_Interp *interp,
			    char *part1, char *part2, int flags,
			    Dcl_VarTraceProc *proc, ClientData clientData));
EXTERN int		Dcl_UpVar _ANSI_ARGS_((Dcl_Interp *interp,
			    char *frameName, char *varName,
			    char *localName, int flags));
EXTERN int		Dcl_UpVar2 _ANSI_ARGS_((Dcl_Interp *interp,
			    char *frameName, char *part1, char *part2,
			    char *localName, int flags));
EXTERN int		Dcl_VarEval _ANSI_ARGS_(VARARGS(Dcl_Interp *interp));
EXTERN ClientData	Dcl_VarTraceInfo _ANSI_ARGS_((Dcl_Interp *interp,
			    char *varName, int flags,
			    Dcl_VarTraceProc *procPtr,
			    ClientData prevClientData));
EXTERN ClientData	Dcl_VarTraceInfo2 _ANSI_ARGS_((Dcl_Interp *interp,
			    char *part1, char *part2, int flags,
			    Dcl_VarTraceProc *procPtr,
			    ClientData prevClientData));

#endif /* _DCL */
